#    Copyright 2009, Bryan Cole, Teraview Ltd.
#
#    This file is part of Raypier.
#
#    Raypier is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

from traits.api import on_trait_change, Float, Instance,Event, Int,\
        Property, Button, Str, Array, List, observe

from traitsui.api import View, Item, VGroup, DropEditor

from tvtk.api import tvtk
import numpy
import time

from raypier.bases import Probe, Traceable, NumEditor, Vector, BaseRayCollection
from raypier.sources import RayCollection, GaussletCollection
from raypier.core.ctracer import FaceList, select_ray_intersections, select_gausslet_intersections#detect_segment, detect_gausslet
from raypier.core.cfaces import RectangularFace

from raypier.utils import normaliseVector

try:
    from raypier.point_spread_func import PSF
except ImportError:
    PSF=None


class BaseCapturePlane(Probe):
    name = Str("Ray Capture Plane")
    width = Float(30.0)
    height = Float(20.0)
    
    face_list = Instance(FaceList)
    
    plane_src = Instance(tvtk.PlaneSource, ())
    
    captured = List(Instance(RayCollection))
    
    vtkproperty = Instance(tvtk.Property, (), {'opacity': 1.0, 'color': (0.1,0.1,0.1),
                                               'representation': 'wireframe'})
    _mtime = Float(0.0)
    
    ### A list of lists of rays. One list per source.
    _all_rays = List()
    
    
    traits_view = View(VGroup(
                       Traceable.uigroup,
                       Item('width', editor=NumEditor),
                       Item('height', editor=NumEditor),
                        ),
                   )
    
    @observe("centre, orientation, width, height")
    def config_pipeline(self, evt):
        hwidth = self.width/2.
        hheight = self.height/2.
        plane = self.plane_src
        plane.origin = (-hwidth, -hheight,0.0)
        plane.point1 = (hwidth, -hheight,0.0)
        plane.point2 = (-hwidth, +hheight,0.0)
        plane.modified()
        
        face = self.face_list.faces[0]
        face.length=self.height
        face.width = self.width
        
        self.update = True
    
    def _actors_default(self):
        plane = self.plane_src
        self.config_pipeline(None)
        
        trans = tvtk.TransformFilter(input_connection = plane.output_port,
                                     transform=self.transform)
        
        map = tvtk.PolyDataMapper(input_connection=trans.output_port)
        map.scalar_visibility = False
        act = tvtk.Actor(mapper=map)
        act.property = self.vtkproperty
        actors = tvtk.ActorCollection()
        actors.append(act)
        return actors
    
    def _face_list_default(self):
        face = RectangularFace(length=self.height, width=self.width, offset=0.0, z_plane=0.0)
        fl = FaceList(owner=self)
        fl.faces = [face,]
        return fl
    
    def evaluate(self, src_list):
        mtime = self._mtime
        if any(src._mtime > mtime for src in src_list):
            all_rays = []
            for src in src_list:
                all_rays.append(src.traced_rays)
            self._all_rays = all_rays
    
    @observe("centre, orientation, width, height, _all_rays")
    def on_input_changed(self, evt):
        self._evaluate()
    
    def _evaluate(self):
        raise NotImplementedError()
        
        
class RayCapturePlane(BaseCapturePlane):
    def _evaluate(self):
        all_rays = self._all_rays
        if not all_rays:
            return
        fl = self.face_list
        fl.sync_transforms()
        captured = [select_ray_intersections(fl, list(rc_list)) for rc_list in all_rays]
        self._mtime = time.monotonic()
        self.captured = captured
        
        
class GaussletCapturePlane(BaseCapturePlane):
    name = Str("Gausslet Capture Plane")
    captured = List(Instance(GaussletCollection))
    
    def _evaluate(self):
        all_rays = self._all_rays
        if not all_rays:
            return
        fl = self.face_list
        fl.sync_transforms()
        captured = [select_gausslet_intersections(fl, list(gc_list)) for gc_list in all_rays]
        self._mtime = time.monotonic()
        self.captured = captured
            

class PolarisationProbe(Probe):
    name = "Polarisation Probe"
    size = Float(25.0)
    
    update = Event() #request re-tracing
    
    plane_src = Instance(tvtk.PlaneSource, (), 
                    {"x_resolution":1, "y_resolution":1},
                    transient=True)
                    
    traits_view = View(VGroup(
                       Traceable.uigroup,
                       Item('size', editor=NumEditor),
                        ),
                   )
    
    @on_trait_change("size")
    def config_pipeline(self):
        src = self.plane_src
        size = self.size/2.
        
        src.origin = (-size,-size,0)
        src.point1 = (-size,size,0)
        src.point2 = (size,-size,0)
        
        self.update = True
        
    @on_trait_change("size, centre, direction")
    def on_change(self):
        self.update = True
    
    def _actors_default(self):
        source = self.plane_src
        trans_f = tvtk.TransformFilter(input_connection=source.output_port,
                        transform = self.transform)
        map = tvtk.PolyDataMapper(input_connection=trans_f.output_port)
        act = tvtk.Actor(mapper=map)
        actors = tvtk.ActorCollection()
        actors.append(act)
        return actors
        
    def get_actors(self, scene):
        return self.actors
        
    def find_intersections(self, ray_src):
        for rays in ray_src.TracedDetailRays:
            self.intersect_rays(rays)
            
    def intersect_rays(self, rays):
        """
        @param rays: a RayCollection instance
        """
        #raise NotImplementedError
        pass
      
if PSF is not None:
    class PointSpreadFunction(Probe):
        position = Vector(desc="centre point of the point grid on which the PSF is evaluated")
        direction = Vector(desc="normal vector of the grid plane")
        orientation = Vector(desc="x_axis direction of the grid plane. This is projected onto\
        the grid plane to get the actual axis direction")
        
        point_spacing = Float(0.05) #in mm
        size = Int(30)
        
        psf = Instance(PSF)
        
        wavelength = Float(100.0, desc="wavelength, in microns")
        
        ray_source = Instance(klass="raypier.sources.BaseRaySource")
        
        exit_pupil_offset = Float(100.0)
        aperture = Float(1.0)
        
        eval_btn = Button("calculate")
        
        def _psf_default(self):
            return PSF(owner=self) 
            
        def _eval_btn_fired(self):
            source = self.ray_source
            
            pass
            
    class FaceCenteredPSF(PointSpreadFunction):
        name = Property(Str, depends_on="target_face")
        position = Property(Vector, depends_on="target_face",
                            desc="centre point of the point grid on which the PSF is evaluated")
        direction = Property(Vector, depends_on="target_face",
                          desc="normal vector of the grid plane")
        orientation = Property(Vector, depends_on="target_face",
                               desc="x_axis direction of the grid plane. This is projected onto\
        the grid plane to get the actual axis direction")
        
        target_face = Instance(klass="raypier.faces.Face")
        
        target_points = Property(Array(shape=(None,3)),
                                  depends_on="position, direction, orientation, size, point_spacing")
        
        rays = Instance(klass="raypier.rays.RayCollection")
        
        source = Instance(tvtk.ProgrammableSource, ())
        
        traits_view = View(Item('eval_btn', show_label=False),
                           VGroup(
                               Item('point_spacing'),
                               Item('size'),
                               Item('exit_pupil_offset'),
                               Item('ray_source', editor=DropEditor()),
                               Item('target_face', editor=DropEditor())
                               ),
                            resizable=True)
        
        def _actors_default(self):
            source = self.source
            def execute():
                output = source.poly_data_output
                if self.rays is None:
                    return
                points = self.rays.origin
                cells = self.rays.cells
                output.points = points
                output.polys = cells.tolist()
                #print "PSF", cells, points
            source.set_execute_method(execute)
            
            map = tvtk.PolyDataMapper(input_connection=source.output_port)
            act = tvtk.Actor(mapper=map)
            act.property.representation="wireframe"
            actors = tvtk.ActorCollection()
            actors.append(act)
            return actors
        
        def _rays_changed(self):
            self.source.modified()
            self.render = True
        
        def _get_name(self):
            optic = self.target_face.owner
            idx = optic.faces.index(self.target_face)
            return "PSF on %s, face %d"%(optic.name, idx)
                           
        def _get_position(self):
            return numpy.asarray(self.target_face.owner.centre)
        
        def _get_direction(self):
            return -numpy.asarray(self.target_face.owner.direction)
        
        def _get_orientation(self):
            return numpy.asarray(self.target_face.owner.x_axis)
        
        def _get_target_points(self):
            length = (self.point_spacing * self.size)/2.
            size = self.size * 1j
            U,V = numpy.ogrid[-length:length:size, -length:length:size]
            
            axis1 = normaliseVector(numpy.cross(self.direction, self.orientation))
            axis2 = normaliseVector(numpy.cross(axis1, self.direction))
            
            newaxis = numpy.newaxis
            points = axis2[newaxis, newaxis,:] * U[:,:,newaxis] \
                    + axis1[newaxis, newaxis,:] * V[:,:,newaxis]
                    
            points += self.position
            return points
            
        
        def _eval_btn_fired(self):
            source = self.ray_source
            
            sequence = source.get_sequence_to_face(self.target_face)
            input_rays = source.InputDetailRays
            
            result = self.psf.do_trace(input_rays, sequence)
            
            self.rays = result
            
            target_points = self.target_points
            
            spot = self.psf.evaluate_scalar_amp(result, target_points)
            
            self.spot = spot



#    Copyright 2009, Teraview Ltd., Bryan Cole
#
#    This file is part of Raypier.
#
#    Raypier is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

#import distribute_setup
#distribute_setup.use_setuptools()
from setuptools import find_packages, setup
#from distutils.core import setup
from distutils.extension import Extension
from Cython.Distutils import build_ext
from Cython.Build import cythonize
#from setuptools import find_packages

###Cython.Distutils and Distribute don't play nice so we'll create the
###C-modules explicitly
import subprocess, os, sys

# def create_module(pyx_name):
#     cname = os.path.splitext(pyx_name)[0] + ".c"
#     if os.path.exists(cname) and \
#         os.stat(cname).st_mtime > os.stat(pyx_name).st_mtime:
#         return
#     ret = subprocess.call(['cython',pyx_name])
#     if ret < 0:
#         raise CythonError("Failed to compile %s with Cython"%pyx_name)

# for fname in ['ctracer.pyx','cfaces.pyx','cmaterials.pyx']:
#     create_module("raypier/%s"%fname)

import numpy
includes = [numpy.get_include()]
libpath = []


Win64 = sys.platform.startswith("win")

# if sys.platform.startswith('win32'):
#     os.environ['PATH'] = os.environ['PATH']+r";C:\Program Files\Microsoft SDKs\Windows\v6.0A\bin"
#     includes.append(r"C:\Program Files\Microsoft SDKs\Windows\v6.0A\Include")
#     libpath += [r"C:\Program Files\Microsoft SDKs\Windows\v6.0A\Lib"]

ext_modules = cythonize("raypier/core/*.pyx",
                        language="c++" if Win64 else None,
                        include_path=[numpy.get_include()],
                        language_level = "3")

for module in ext_modules:
    module.extra_compile_args.append('-fopenmp')
    module.extra_link_args.append('-fopenmp')
    
with open("README.rst", 'r', encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="raypier",
    version="0.2.1",
    packages=find_packages(),
    scripts = [], #no stand-alone application yet
    cmdclass = {'build_ext': build_ext},
    #ext_package = "raypier", ###Not needed on linux. Is it necessary on Win64?
    ext_modules = ext_modules,
    include_dirs = [numpy.get_include()],
    zip_safe = False, #Zipped eggs cause the ReadTheDocs build to fail.

    install_requires = [
        #"VTK",
        #"wxPython", #maybe can avoid an explicit dependancy on wx
        "numpy == 1.19", 
        "traits >= 3.0",
        "mayavi >= 3.0", #TVTK is distributed as part of Mayavi
        "traitsui >= 3.0"
        ],

    package_data = {
        "": ["*.pyx", "*.pxd"],
        "raypier": ["material_data/glass_dispersion_database.db"]
        },

    author = "Bryan Cole",
    author_email = "bryancole.cam@gmail.com",
    description = """A optical ray-tracing package, for design, optimisation and \
visualisation of mirror/lens systems.""",
    long_description = long_description,
    license = "GPL3",
    keywords = "science engineering optics ray-tracing physcics",
    classifiers=["Development Status :: 4 - Beta",
                 "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
                 "Natural Language :: English",
                 "Programming Language :: Cython",
                 "Programming Language :: Python :: 3.7",
                 "Programming Language :: Python :: Implementation :: CPython",
                 "Topic :: Scientific/Engineering :: Physics",
                 "Topic :: Scientific/Engineering :: Visualization"
                 ],
    url = "https://groups.google.com/u/1/g/python-raytrace",
    project_urls = {
        "Homepage" : "https://groups.google.com/u/1/g/python-raytrace",
        "Documentation" : "https://raypier-optics.readthedocs.io/en/latest/index.html",
        "Source" : "https://github.com/bryancole/raypier_optics.git",
        "Issues" : "https://github.com/bryancole/raypier_optics/issues"
        },
    python_requires=">=3.7"
    )



from traits.api import Float, Instance, on_trait_change
from traitsui.api import View, Item, ListEditor, VSplit,\
            RangeEditor, ScrubberEditor, HSplit, VGroup
from tvtk.api import tvtk

from raypier.bases import NumEditor, Traceable
from raypier.mirrors import BaseMirror
from raypier.core.cfaces import OffAxisParabolicFace
from raypier.core.ctracer import FaceList

from raypier.vtk_algorithms import EmptyGridSource


class OffAxisParabloid(BaseMirror):
    name = "Off Axis Parabolic"
    abstract = False
    diameter = Float(25.0)
    EFL = Float(50.0, desc="Effective Focal Length")
    height = Float(50.0)
    
    max_length=Float(200.0) #what does this do?
    
    vtk_grid = Instance(EmptyGridSource, ())
    vtk_cylinder = Instance(tvtk.Cylinder, ())
    vtk_quadric = Instance(tvtk.Quadric, ())
    
    traits_view = View(VGroup(
                        Traceable.uigroup,
                       Item('height', editor=NumEditor),
                       Item('diameter', editor=NumEditor),
                       Item('EFL', editor=NumEditor),
                       Item('max_length', editor=NumEditor),
                       ),
                       )
    
    def _faces_default(self):
        facelist = FaceList(owner=self)
        facelist.faces=[OffAxisParabolicFace(owner=self,
                                             EFL=self.EFL,
                                             diameter=self.diameter)]
        return facelist
    
    @on_trait_change("EFL, diameter")
    def config_pipeline(self):
        face = self.faces.faces[0]
        face.EFL = EFL = self.EFL
        face.diameter = self.diameter
        
        rad = self.diameter/2
        h = self.height
        
        self.update_grid()
        self.vtk_grid.modified()
                      
        cyl = self.vtk_cylinder
        cyl.center = (EFL,0,0)
        cyl.radius = rad
        
        q = self.vtk_quadric
        A0 = A1 = -1/(2*EFL)
        A8 = 1 
        A9 = EFL/2
        q.coefficients = (A0, A1, 0, 
                          0, 0, 0, 
                          0, 0, A8, 
                          A9)
        self.update=True
        
    def update_grid(self):
        EFL = self.EFL
        r = self.diameter/2
        h = self.height
        l = self.max_length
        source = self.vtk_grid
        size = 20
        spacing = 2*r / (size-1)
        lsize = int(l/spacing)
        source.dimensions = (size,size,lsize)
        source.origin = (EFL - r, -r, -h)
        source.spacing = (spacing, spacing, spacing)
        
    def _pipeline_default(self):
        grid = self.vtk_grid
        self.update_grid()
        grid.modified()
        
        trans = tvtk.Transform()
        trans.rotate_x(90.)
        cyl = self.vtk_cylinder
        cyl.transform = trans
        
        clip1 = tvtk.ClipVolume(input_connection=grid.output_port,
                                 clip_function=self.vtk_cylinder,
                                 inside_out=1)
        
        clip2 = tvtk.ClipDataSet(input_connection = clip1.output_port,
                                 clip_function=self.vtk_quadric,
                                 inside_out=1)
        
        topoly = tvtk.GeometryFilter(input_connection=clip2.output_port)
        norms = tvtk.PolyDataNormals(input_connection=topoly.output_port)
        
        transF = tvtk.TransformFilter(input_connection=norms.output_port, transform=self.transform)
        self.config_pipeline()
        #clip1.update()
        grid.modified()
        return transF


# cython: cdivision=True
cimport cython

cdef extern from "math.h":
    double sqrt(double arg)
    double fabs(double arg)
    double sin(double arg)
    double cos(double arg)
    double atan2(double y, double x)
    double erf(double arg)
    double M_PI
    
cdef extern from "float.h":
    double DBL_MAX

IF UNAME_SYSNAME == "Windows":
    cdef extern from "complex.h":
        double complex csqrt "sqrt" (double complex)
        double cabs "abs" (double complex)
        double complex cexp "exp" (double complex)
	
    cdef double complex I = 1j
ELSE:
    cdef extern from "complex.h":
        double complex csqrt (double complex)
        double cabs (double complex)
        double complex cexp (double complex)
        double complex I
    
cdef:
    double INF=(DBL_MAX+DBL_MAX)
    double SP_TOL=1.0e-10

from .ctracer cimport InterfaceMaterial, norm_, dotprod_, \
        multvs_, subvv_, vector_t, ray_t, RayCollection, \
        complex_t, mag_sq_, Ray, cross_, set_v, ray_power_, \
        orientation_t, addvv_, REFL_RAY, PARABASAL, \
        para_t, gausslet_t, GaussletCollection

import numpy as np
cimport numpy as np_


cdef ray_t convert_to_sp(ray_t ray, vector_t normal):
    """Project the E-field components of a given ray
    onto the S- and P-polarisations defined by the 
    surface normal
    """
    cdef:
        vector_t E2_vector, E1_vector, v, S_vector, P_vector
        complex_t S_amp, P_amp, E1_amp, E2_amp
        double A, B
    
    E1_amp = ray.E1_amp
    E2_amp = ray.E2_amp
    E2_vector = norm_(cross_(ray.direction, ray.E_vector))
    E1_vector = norm_(cross_(E2_vector, ray.direction))
    normal = norm_(normal)
    #print("E_in:", E1_amp.real, E1_amp.imag, E2_amp.real, E2_amp.imag)
    #print("E_vector:", E1_vector.x, E1_vector.y, E1_vector.z, E2_vector.x, E2_vector.y, E2_vector.z)
    S_vector = cross_(ray.direction, normal)
    if fabs(S_vector.x)<SP_TOL and fabs(S_vector.y)<SP_TOL and fabs(S_vector.z)<SP_TOL:
        return ray
    S_vector = norm_(S_vector)
        
    v = cross_(ray.direction, S_vector)
    P_vector = norm_(v)
    
    A = dotprod_(E1_vector, S_vector)
    B = dotprod_(E2_vector, S_vector)
    
    S_amp.real = E1_amp.real*A + E2_amp.real*B
    S_amp.imag = E1_amp.imag*A + E2_amp.imag*B
    
    B = dotprod_(E1_vector, P_vector)
    A = dotprod_(E2_vector, P_vector)

    P_amp.real = E1_amp.real*B + E2_amp.real*A
    P_amp.imag = E1_amp.imag*B + E2_amp.imag*A
    
    ray.E_vector = S_vector
    ray.E1_amp = S_amp
    ray.E2_amp = P_amp
    
    #ray.phase = 0.0
    return ray


def Convert_to_SP(Ray ray, normal):
    cdef vector_t n
    cdef ray_t r
    n = set_v(normal)
    r = convert_to_sp(ray.ray, n)
    out = Ray()
    out.ray = r
    return out


ctypedef double (*dispersion_curve) (double, double[:])


cdef double nondispersive_0(double wavelen, double[:] coefs):
    return coefs[0]

cdef double sellmeier_1(double wavelen, double[:] coefs ):
    cdef:
        double n2=1.0
        double wl2=wavelen*wavelen
        int i, n_coefs = coefs.shape[0]
        
    n2 += coefs[0]
    for i in range((n_coefs-1)/2):
        n2 += coefs[2*i + 1] * wl2 / (wl2 - coefs[2*i + 2]**2)
    return sqrt(n2)

cdef double sellmeier_2(double wavelen, double[:] coefs ):
    cdef:
        double n2=1.0
        double wl2=wavelen*wavelen
        int i, n_coefs = coefs.shape[0]
        
    n2 += coefs[0]
    for i in range((n_coefs-1)/2):
        n2 += coefs[2*i + 1] * wl2 / (wl2 - coefs[2*i + 2])
    return sqrt(n2)

### Polynomia formula
cdef double sellmeier_3(double wavelen, double[:] coefs ):
    cdef:
        double n2=coefs[0]
        int i, n_coefs = coefs.shape[0]
        
    for i in range(1,n_coefs-1,2):
        n2 += coefs[i]*( wavelen**(coefs[i+1]) )
    return sqrt(n2)
        

cdef double sellmeier_5(double wavelen, double[:] coefs):
    cdef:
        double n= coefs[0]
        int i
        
    for i in range((coefs.shape[0]-1)/2):
        n += coefs[2*i + 1] * (wavelen**coefs[2*i+2])
        
        
cdef class BaseDispersionCurve(object):
    """
    Base class for DispersionCurve objects. This extension class provides 
    efficient C-API methods for evaluation of the refractive index for a
    given wavelength, based on one of various dispersion functions (e.g. 
    Sellmeier curves) and a set of coefficients obtained from https://refractiveindex.info
    
    :param formula_id: Specifies the formula to use in evaluating this dispersion-curve
    :type formula_id: int, readonly
    :param wavelength_min: Minimum wavelength that can be evaluated, in microns.
    :type wavelength_min: float, readonly
    :param wavelength_max: Maximum wavelength that can be evaluated, in microns.
    :type wavelength_max: float, readonly
    :param coefs: an array-like list of coefficients for the particular formula used.
    :type coefs: list[float], readonly
    :param absorption: The (constant) absorption coefficient used to evaluate the complex refractive index.
                        Given in cm^-1.
    :type absorption: float, readonly
    """
    cdef:
        readonly double wavelength_min
        readonly double wavelength_max
        readonly double[:] coefs
        readonly double absorption #in cm^-1
        readonly int formula_id
        dispersion_curve curve
        
    def __init__(self, int formula_id, double[:] coefs, 
                  double absorption=0.0,
                  double wavelength_min=0.1, 
                  double wavelength_max=1000000.0):
        """
        Object constructor
        
        :param int formula_id: The formula identifier {0,1,2,3,5}
        """
        self.coefs = coefs
        if formula_id==1:
            self.curve = &sellmeier_1
        elif formula_id==2:
            self.curve = &sellmeier_2
        elif formula_id==3:
            self.curve = &sellmeier_3
        elif formula_id==5:
            self.curve = &sellmeier_5
        elif formula_id==0:
            self.curve = &nondispersive_0
        else:
            raise ValueError("Unknown formula id (%d)"%(formula_id,))
        self.formula_id = formula_id
        self.absorption = absorption
        self.wavelength_min = wavelength_min
        self.wavelength_max = wavelength_max
        
    cdef np_.npy_complex128[:] c_evaluate_n(self, double[:] wavelen):
        cdef:
            dispersion_curve curve=self.curve
            double[:] coefs=self.coefs
            int i
            np_.npy_complex128[:] n_out
            double n_imag #imaginary part of n due to absorption
            double wvl
        
        n_imag = 0.0001 * self.absorption / (4* M_PI)
        
        n_out = np.empty(wavelen.shape[0], dtype=np.complex128)
        for i in range(wavelen.shape[0]):
            wvl = wavelen[i] #in microns
            if (wvl < self.wavelength_min) or (wvl > self.wavelength_max):
                msg = "Wavelength (%f) outside range of dispersion curve (%f -> %f)"%(wvl, 
                                                                                      self.wavelength_min,
                                                                                      self.wavelength_max)
                raise ValueError(msg) 
            n_out[i].real = curve(wvl, coefs)
            n_out[i].imag = n_imag * wvl
            
        return n_out
    
    def evaluate_n(self, wavelen):
        """
        Calculates the complex refractive index for the given wavelengths.
        
        :param wavelen: An array-like collection of wavelength, in microns.
        :type wavelen: double[:]
        :return: The refractive index for the given wavelength.
        :rtype: complex128[:]
        """
        return np.asarray(self.c_evaluate_n(np.asarray(wavelen)))


cdef class OpaqueMaterial(InterfaceMaterial):
    """A perfect absorber i.e. it generates no rays
    """
    cdef void eval_child_ray_c(self,
                            ray_t *in_ray, 
                            unsigned int idx, 
                            vector_t point,
                            orientation_t orient,
                            RayCollection new_rays):
        pass
    
    
cdef class TransparentMaterial(InterfaceMaterial):
    """A perfect transmitter i.e. it generates an
    outgoing ray with identical direction, polarisation etc.
    to the incoming ray. It does project the polarisation
    vectors to it's S- and P-directions, however.
    """
    cdef void eval_child_ray_c(self,
                            ray_t *in_ray, 
                            unsigned int idx, 
                            vector_t point,
                            orientation_t orient,
                            RayCollection new_rays):
        cdef:
            vector_t normal
            ray_t sp_ray
        
        normal = norm_(orient.normal)
        sp_ray = convert_to_sp(in_ray[0], normal)
        sp_ray.accumulated_path += sp_ray.length * sp_ray.refractive_index.real
        sp_ray.origin = point
        sp_ray.normal = normal
        sp_ray.length = INF
        sp_ray.parent_idx = idx
        sp_ray.ray_type_id &= ~REFL_RAY
        new_rays.add_ray_c(sp_ray)


cdef class PECMaterial(InterfaceMaterial):
    """Simulates a Perfect Electrical Conductor. I.e. incident rays are reflected with 
    100% reflectivity.
    """
    cdef void eval_child_ray_c(self,
                            ray_t *in_ray, 
                            unsigned int idx, 
                            vector_t point,
                            orientation_t orient,
                            RayCollection new_rays):
        """
           ray - the ingoing ray
           idx - the index of ray in it's RayCollection
           point - the position of the intersection (in global coords)
           normal - the outward normal vector for the surface
        """
        cdef:
            vector_t cosThetaNormal, reflected, normal
            ray_t sp_ray
            complex_t cpx
            double cosTheta
        
        normal = norm_(orient.normal)
        sp_ray = convert_to_sp(in_ray[0], normal)
        cosTheta = dotprod_(normal, in_ray.direction)
        cosThetaNormal = multvs_(normal, cosTheta)
        reflected = subvv_(in_ray.direction, multvs_(cosThetaNormal, 2))
        sp_ray.accumulated_path += sp_ray.length * sp_ray.refractive_index.real
        sp_ray.origin = point
        sp_ray.normal = normal
        sp_ray.direction = reflected
        sp_ray.length = INF
        sp_ray.E1_amp.real = -sp_ray.E1_amp.real
        sp_ray.E1_amp.imag = -sp_ray.E1_amp.imag
        #sp_ray.E2_amp.real = -sp_ray.E2_amp.real
        #sp_ray.E2_amp.imag = -sp_ray.E2_amp.imag
        sp_ray.parent_idx = idx
        sp_ray.ray_type_id |= REFL_RAY
        new_rays.add_ray_c(sp_ray)
        
        
cdef class PartiallyReflectiveMaterial(InterfaceMaterial):
    """
    A simple materials with a fixed reflectivity.
    
    :param reflectivity: The material power-reflectivity given as a value from 0.0. to 1.0 
    :type reflectivity: double
    """
    cdef:
        ### power reflectivity, from 0. to 1.
        public double _reflectivity
        
    def __cinit__(self, **kwds):
        self.reflectivity = kwds.get("reflectivity", 0.5)
        
    property reflectivity:
        def __get__(self):
            return self._reflectivity
        
        def __set__(self, val):
            if val<0.0 or val>1.0:
                raise ValueError(f"Reflectivity must be in range 0.0 to 1.0 ({val} given).")
            self._reflectivity = val
        
    cdef void eval_child_ray_c(self,
                            ray_t *in_ray, 
                            unsigned int idx, 
                            vector_t point,
                            orientation_t orient,
                            RayCollection new_rays):
        """
           ray - the ingoing ray
           idx - the index of ray in it's RayCollection
           point - the position of the intersection (in global coords)
           normal - the outward normal vector for the surface
        """
        cdef:
            vector_t cosThetaNormal, reflected, normal
            vector_t tangent, tg2, in_direction
            ray_t sp_ray, sp_ray2
            double cosTheta, R, T
            complex_t P
            
        normal = norm_(orient.normal)
        in_direction = norm_(in_ray.direction)
        sp_ray = sp_ray2 = convert_to_sp(in_ray[0], normal)
        cosTheta = dotprod_(normal, in_direction)            
        cosThetaNormal = multvs_(normal, cosTheta)
        
        sp_ray.accumulated_path += sp_ray.length * sp_ray.refractive_index.real
        sp_ray2.accumulated_path = sp_ray.accumulated_path
        
        R = sqrt(self._reflectivity)
        T = sqrt(1-self._reflectivity)
        
        #Reflected ray
        reflected = subvv_(in_direction, multvs_(cosThetaNormal, 2))
        sp_ray.origin = point
        sp_ray.normal = normal
        sp_ray.direction = reflected
        sp_ray.length = INF
        sp_ray.E1_amp *= R
        sp_ray.E2_amp *= R
        sp_ray.parent_idx = idx
        sp_ray.ray_type_id |= REFL_RAY
        new_rays.add_ray_c(sp_ray)
            
        #Transmitted ray
        sp_ray2.origin = point
        sp_ray2.normal = normal
        sp_ray2.direction = in_direction
        sp_ray2.length = INF
        sp_ray2.E1_amp *= T
        sp_ray2.E2_amp *= T
        sp_ray2.parent_idx = idx
        sp_ray2.ray_type_id &= ~REFL_RAY
        new_rays.add_ray_c(sp_ray2)
        
        
cdef class LinearPolarisingMaterial(InterfaceMaterial):
    """Simulates a perfect polarising beam splitter. P-polarisation
    is 100% transmitted while S- is reflected"""
    
    cdef void eval_child_ray_c(self,
                            ray_t *in_ray, 
                            unsigned int idx, 
                            vector_t point,
                            orientation_t orient,
                            RayCollection new_rays):
        """
           ray - the ingoing ray
           idx - the index of ray in it's RayCollection
           point - the position of the intersection (in global coords)
           normal - the outward normal vector for the surface
        """
        cdef:
            vector_t cosThetaNormal, reflected, normal
            vector_t tangent, tg2, in_direction
            ray_t sp_ray, sp_ray2
            double cosTheta
            complex_t P
            
        normal = norm_(orient.normal)
        in_direction = norm_(in_ray.direction)
        sp_ray = sp_ray2 = convert_to_sp(in_ray[0], normal)
        cosTheta = dotprod_(normal, in_direction)            
        cosThetaNormal = multvs_(normal, cosTheta)
        
        sp_ray.accumulated_path += sp_ray.length * sp_ray.refractive_index.real
        sp_ray2.accumulated_path = sp_ray.accumulated_path
        
        #Reflect the S-polarisation
        reflected = subvv_(in_direction, multvs_(cosThetaNormal, 2))
        sp_ray.origin = point
        sp_ray.normal = normal
        sp_ray.direction = reflected
        sp_ray.length = INF
        #set P-polarisation to zero
        sp_ray.E2_amp.real = 0.0
        sp_ray.E2_amp.imag = 0.0
        sp_ray.parent_idx = idx
        sp_ray.ray_type_id |= REFL_RAY
        new_rays.add_ray_c(sp_ray)
            
        #Transmit the P-polarisation
        sp_ray2.origin = point
        sp_ray2.normal = normal
        sp_ray2.direction = in_direction
        sp_ray2.length = INF
        #set S-polarisation to zero
        sp_ray2.E1_amp.real = 0.0
        sp_ray2.E1_amp.imag = 0.0
        sp_ray2.parent_idx = idx
        sp_ray2.ray_type_id &= ~REFL_RAY
        new_rays.add_ray_c(sp_ray2)
        
        
cdef class WaveplateMaterial(InterfaceMaterial):
    """An idealised optical retarder.
    
    :param retardance: The optical retardance, given in terms of numbers-of-wavelengths.
    :type retardance: double
    :param fast_axis: A vector giving the "fast" polarisation axis
    :type fast_axis: (double, double, double)
    """
    cdef complex_t retardance_
    cdef vector_t fast_axis_
    
    property retardance:
        def __get__(self):
            cdef double val
            val = atan2(self.retardance_.imag, 
                         self.retardance_.real)/(2*M_PI)
            if val < 0:
                val += 1.0
            return val
            
        def __set__(self, val):
            self.retardance_.real = cos(val*2*M_PI)
            self.retardance_.imag = sin(val*2*M_PI)
    
    property fast_axis:
        def __get__(self):
            cdef vector_t ax = self.fast_axis_
            return (ax.x, ax.y, ax.z)
        
        def __set__(self, ax):
            self.fast_axis_.x = ax[0]
            self.fast_axis_.y = ax[1]
            self.fast_axis_.z = ax[2]
            
            
    def __cinit__(self, **kwds):
        self.retardance = kwds.get("retardance", 0.25)
        self.fast_axis = kwds.get("fast_axis", (1.0,0,0))
            
            
    cdef ray_t apply_retardance_c(self, ray_t r):
        cdef:
            complex_t E1=r.E1_amp, retard=self.retardance_
        r.E1_amp.real = E1.real*retard.real - E1.imag*retard.imag
        r.E1_amp.imag = E1.imag*retard.real + E1.real*retard.imag
        #Leave E2_amp untouched.
        return r
    
    def apply_retardance(self, Ray r):
        """Applies the retardance to the given Ray object.
        
        :param r: an input Ray object.
        :type r: raypier.core.ctracer.Ray
        :return: a new Ray instance
        :rtype: raypier.core.ctracer.Ray
        """
        cdef Ray out
        out = Ray()
        out.ray = self.apply_retardance_c(r.ray)
        return out
            
    
    cdef void eval_child_ray_c(self,
                            ray_t *in_ray, 
                            unsigned int idx, 
                            vector_t point,
                            orientation_t orient,
                            RayCollection new_rays):
        """
           ray - the ingoing ray
           idx - the index of ray in it's RayCollection
           point - the position of the intersection (in global coords)
           normal - the outward normal vector for the surface
        """
        cdef:
            vector_t normal
            ray_t out_ray
            complex_t E1, retard
            
        normal = norm_(orient.normal)
        in_direction = norm_(in_ray.direction)
        ###The "P" output of convert_to_sp with be aligned with the fast_axis
        ###The "S" output will thus be orthogonal to the fast axis
        out_ray = convert_to_sp(in_ray[0], self.fast_axis_)
        out_ray.accumulated_path += out_ray.length * out_ray.refractive_index.real
        out_ray.origin = point
        out_ray.normal = normal
        out_ray.direction = in_direction
        out_ray.length = INF
        out_ray.parent_idx = idx
        out_ray.ray_type_id &= ~REFL_RAY
        out_ray = self.apply_retardance_c(out_ray)
        
        new_rays.add_ray_c(out_ray)
    
    
cdef class DielectricMaterial(InterfaceMaterial):
    """Simulates Fresnel reflection and refraction at a
    normal dielectric interface.
    
    The surface normal is assumed to be pointing "out" of the material.
    
    :param complex n_inside: The refractive index on the inside of the material interface
    :param complex n_outside: The refractive index on the outside of the material interface
    """
    cdef complex_t n_inside_, n_outside_
    
    def __cinit__(self, **kwds):
        self.n_inside = kwds.get('n_inside', 1.5)
        self.n_outside = kwds.get('n_outside', 1.0)
    
    property n_inside:
        def __get__(self):
            return complex(self.n_inside_.real, self.n_inside_.imag)
        
        def __set__(self, v):
            v = complex(v)
            self.n_inside_.real = v.real
            self.n_inside_.imag = v.imag
            
    property n_outside:
        def __get__(self):
            return complex(self.n_outside_.real, self.n_outside_.imag)
        
        def __set__(self, v):
            v = complex(v)
            self.n_outside_.real = v.real
            self.n_outside_.imag = v.imag
            
    cdef void eval_child_ray_c(self,
                            ray_t *in_ray, 
                            unsigned int idx, 
                            vector_t point,
                            orientation_t orient,
                            RayCollection new_rays):
        """
           ray - the ingoing ray
           idx - the index of ray in it's RayCollection
           point - the position of the intersection (in global coords)
           normal - the outward normal vector for the surface
        """
        cdef:
            vector_t cosThetaNormal, reflected, transmitted
            vector_t tangent, tg2, in_direction, normal
            ray_t sp_ray
            complex_t cpx
            double cosTheta, n1, n2, N2, cos1
            double N2cosTheta, N2_sin2, tan_mag_sq, c2
            double cos2, Two_n1_cos1, aspect, T_p, T_s
            int flip
            
        normal = norm_(orient.normal)
        in_direction = norm_(in_ray.direction)
        sp_ray = convert_to_sp(in_ray[0], normal)
        sp_ray.accumulated_path += sp_ray.length * sp_ray.refractive_index.real
        cosTheta = dotprod_(normal, in_direction)
        cos1 = fabs(cosTheta)
        
        #print "TRACE"
        #print normal, in_direction
        
        if cosTheta < 0.0: 
            #ray incident from outside going inwards
            n1 = self.n_outside_.real
            n2 = self.n_inside_.real
            sp_ray.refractive_index = self.n_inside_
            flip = 1
            #print "out to in", n1, n2
        else:
            n1 = self.n_inside_.real
            n2 = self.n_outside_.real
            sp_ray.refractive_index = self.n_outside_
            flip = -1
            #print "in to out", n1, n2
            
        N2 = (n2/n1)**2
        N2cosTheta = N2*cos1
        
        N2_sin2 = (cosTheta*cosTheta) + (N2 - 1)
        #print "TIR", N2_sin2, cosTheta, N2, cos1
        #print (normal.x, normal.y, normal.z), in_direction
        cosThetaNormal = multvs_(normal, cosTheta)
        if N2_sin2 < 0.0:
            #total internal reflection
            reflected = subvv_(in_direction, multvs_(cosThetaNormal, 2))
            sp_ray.origin = point
            sp_ray.normal = normal
            sp_ray.direction = reflected
            sp_ray.length = INF
            sp_ray.E1_amp.real *= -1
            sp_ray.E1_amp.imag *= -1
            sp_ray.E2_amp.real *= -1
            sp_ray.E2_amp.imag *= -1
            sp_ray.parent_idx = idx
            sp_ray.ray_type_id |= REFL_RAY
        else:
            #normal transmission            
            tangent = subvv_(in_direction, cosThetaNormal)
            tg2 = multvs_(tangent, n1/n2)
            tan_mag_sq = mag_sq_(tg2)
            c2 = sqrt(1-tan_mag_sq)
            transmitted = subvv_(tg2, multvs_(normal, c2*flip))
            
            cos2 = fabs(dotprod_(transmitted, normal))
            Two_n1_cos1 = (2*n1)*cos1
            aspect = sqrt(cos2/cos1) * Two_n1_cos1
            
            #Fresnel equations for transmission
            T_p = aspect / ( n2*cos1 + n1*cos2 )
            T_s = aspect / ( n2*cos2 + n1*cos1 )
            #print "T_s", T_s, "T_p", T_p
            
            sp_ray.origin = point
            sp_ray.normal = normal
            sp_ray.direction = transmitted
            sp_ray.length = INF
            sp_ray.E1_amp.real *= T_s
            sp_ray.E1_amp.imag *= T_s
            sp_ray.E2_amp.real *= T_p
            sp_ray.E2_amp.imag *= T_p
            sp_ray.parent_idx = idx
            sp_ray.ray_type_id &= ~REFL_RAY
        new_rays.add_ray_c(sp_ray)
        
        
    cdef para_t eval_parabasal_ray_c(self,
                                     ray_t *base_ray, 
                                     vector_t direction, #incoming ray direction
                                   vector_t point, #position of intercept
                                   orientation_t orient,
                                   unsigned int ray_type_id, #bool, if True, it's a reflected ray
                                   ):
        cdef:
            vector_t cosThetaNormal, normal, out_dir
            para_t para_out
            double cosTheta, n1, n2
            int flip
        
        normal = norm_(orient.normal)
        direction = norm_(direction)
        cosTheta = dotprod_(normal, direction)
        cosThetaNormal = multvs_(normal, cosTheta)
        
        if cosTheta < 0.0: 
            #ray incident from outside going inwards
            n1 = self.n_outside_.real
            n2 = self.n_inside_.real
            flip = 1
        else:
            n1 = self.n_inside_.real
            n2 = self.n_outside_.real
            flip = -1
        
        if ray_type_id & REFL_RAY:
            out_dir = subvv_(direction, multvs_(cosThetaNormal, 2))
        else:
            tangent = subvv_(direction, cosThetaNormal)
            tg2 = multvs_(tangent, n1/n2)
            tan_mag_sq = mag_sq_(tg2)
            c2 = sqrt(1-tan_mag_sq)
            out_dir = subvv_(tg2, multvs_(normal, c2*flip))
        
        para_out.direction = out_dir 
        para_out.origin = point
        para_out.normal = normal
        para_out.length = INF
        return para_out
        

cdef class FullDielectricMaterial(DielectricMaterial):
    """Model for dielectric using full Fresnel equations 
    to give true phase and amplitude response
    
    :param double reflection_threshold: Sets the amplitude threshold for generating a reflected ray.
    :param double transmission_threshold: Sets the amplitude threshold for generating a transmitted ray.
    """
    cdef:
        public double reflection_threshold, transmission_threshold
        public double thickness #in microns
        complex_t n_coating_
        
    property n_coating:
        def __get__(self):
            return complex(self.n_coating_.real, self.n_coating_.imag)
        
        def __set__(self, v):
            v = complex(v)
            self.n_coating_.real = v.real
            self.n_coating_.imag = v.imag
    
    def __cinit__(self, **kwds):
        self.n_coating = kwds.get("n_coating", 1.0)
        self.thickness = kwds.get("thickness", 0.1)
        self.reflection_threshold = kwds.get('reflection_threshold', 0.1)
        self.transmission_threshold = kwds.get('transmission_threshold', 0.1)
    
    @cython.cdivision(True)
    cdef void eval_child_ray_c(self,
                            ray_t *in_ray, 
                            unsigned int idx, 
                            vector_t point,
                            orientation_t orient,
                            RayCollection new_rays):
        """
           ray - the ingoing ray
           idx - the index of ray in it's RayCollection
           point - the position of the intersection (in global coords)
           normal - the outward normal vector for the surface
        """
        cdef:
            vector_t cosThetaNormal, reflected, transmitted
            vector_t tangent, tg2, in_direction, normal
            ray_t sp_ray
            complex_t cpx
            double complex n1, n2, cos2, sin2, R_p, R_s, T_p, T_s, E1_amp, E2_amp
            double cosTheta, cos1, sin1, P_in
            double tan_mag_sq, c2
            double Two_n1_cos1, aspect
            int flip
            
        normal = norm_(orient.normal)
        in_direction = norm_(in_ray.direction)
        #print "in fields:", in_ray[0].E1_amp.real, in_ray[0].E1_amp.imag, in_ray[0].E2_amp.real, in_ray[0].E2_amp.imag
        sp_ray = convert_to_sp(in_ray[0], normal)
        sp_ray.accumulated_path += sp_ray.length * sp_ray.refractive_index.real
        E1_amp = sp_ray.E1_amp.real + 1.0j*sp_ray.E1_amp.imag
        E2_amp = sp_ray.E2_amp.real + 1.0j*sp_ray.E2_amp.imag
        cosTheta = dotprod_(normal, in_direction)
        cos1 = fabs(cosTheta)
        sin1 = sqrt(1 - cos1*cos1)
        
        #print "TRACE"
        #print normal, in_direction
        
        if cosTheta < 0.0: 
            #ray incident from outside going inwards
            n1 = self.n_outside_.real + 1.0j*self.n_outside_.imag
            n2 = self.n_inside_.real + 1.0j*self.n_inside_.imag
            flip = 1
            #print "out to in", n1, n2
        else:
            n1 = self.n_inside_.real + 1.0j*self.n_inside_.imag
            n2 = self.n_outside_.real + 1.0j*self.n_outside_.imag
            flip = -1
            #print "in to out", n1, n2
            
        #apply Snell's law. These become complex.
        sin2 = (n1*sin1/n2)
        cos2 = csqrt(1 - sin2*sin2)
        #print "cos1", cos1
        #print "cos2", cos2
        
        #print "TIR", N2_sin2, cosTheta, N2, cos1
        #print (normal.x, normal.y, normal.z), in_direction
        cosThetaNormal = multvs_(normal, cosTheta)
        
        #incoming power
        P_in =  n1.real*(E1_amp.real**2 + E1_amp.imag**2 + \
                         E2_amp.real**2 + E2_amp.imag**2)
        #print "P_in:", P_in, n1.real, E1_amp.real, E1_amp.imag, E2_amp.real, E2_amp.imag
        if P_in==0.0:
            return
        #print "P Incoming:", P_in
        #Fresnel equations for reflection
        R_p = -(n2*cos1 - n1*cos2)/(n2*cos1 + n1*cos2)
        R_s = -(n2*cos2 - n1*cos1)/(n2*cos2 + n1*cos1)
        
        #modify in place to get reflected amplitudes
        R_s *= E1_amp
        R_p *= E2_amp
        
        #print "P_R:", (cabs(R_s)**2 + cabs(R_p)**2)
        
        if ( n1.real*(cabs(R_s)**2 + cabs(R_p)**2)/P_in ) > self.reflection_threshold:
            reflected = subvv_(in_direction, multvs_(cosThetaNormal, 2))
            sp_ray.origin = point
            sp_ray.normal = normal
            sp_ray.direction = reflected
            sp_ray.length = INF
            sp_ray.E1_amp.real = R_s.real
            sp_ray.E1_amp.imag = R_s.imag
            sp_ray.E2_amp.real = -R_p.real
            sp_ray.E2_amp.imag = -R_p.imag
            sp_ray.parent_idx = idx
            sp_ray.refractive_index.real = n1.real
            sp_ray.refractive_index.imag = n1.imag
            sp_ray.ray_type_id |= REFL_RAY
            new_rays.add_ray_c(sp_ray)
            
        #normal transmission            
        tangent = subvv_(in_direction, cosThetaNormal)
        tg2 = multvs_(tangent, n1.real/n2.real) #This is an approximation for complex N
        tan_mag_sq = mag_sq_(tg2)
        c2 = sqrt(1-tan_mag_sq)
        transmitted = subvv_(tg2, multvs_(normal, c2*flip))
        
        aspect = sqrt(cos2.real/cos1)
        
        #Fresnel equations for transmission
        T_p = aspect * (2.0*cos1*n1) / ( n2*cos1 + n1*cos2 )
        T_s = aspect * (2.0*cos1*n1) / ( n2*cos2 + n1*cos1 )
        #print "T_s", T_s, "T_p", T_p
        
        #modify in place to get reflected amplitudes
        T_s *= E1_amp
        T_p *= E2_amp
        
        if ( n2.real*(cabs(T_s)**2 + cabs(T_p)**2)/P_in ) > self.transmission_threshold:
            sp_ray.origin = point
            sp_ray.normal = normal
            sp_ray.direction = transmitted
            sp_ray.length = INF
            sp_ray.E1_amp.real = T_s.real
            sp_ray.E1_amp.imag = T_s.imag
            sp_ray.E2_amp.real = T_p.real
            sp_ray.E2_amp.imag = T_p.imag
            sp_ray.parent_idx = idx
            sp_ray.refractive_index.real = n2.real
            sp_ray.refractive_index.imag = n2.imag
            sp_ray.ray_type_id &= ~REFL_RAY
            new_rays.add_ray_c(sp_ray)
            
            
cdef class SingleLayerCoatedMaterial(FullDielectricMaterial):
    """
    A material with a single-layer dielectric coating at the interface.
    
    :param complex n_coating: The complex refractive index for the coating.
    :param double thickness: The thickness of the coating in microns
    """
            
    @cython.cdivision(True)
    cdef void eval_child_ray_c(self,
                            ray_t *in_ray, 
                            unsigned int idx, 
                            vector_t point,
                            orientation_t orient,
                            RayCollection new_rays):
        """
           ray - the ingoing ray
           idx - the index of ray in it's RayCollection
           point - the position of the intersection (in global coords)
           normal - the outward normal vector for the surface
        """
        cdef:
            vector_t cosThetaNormal, reflected, transmitted
            vector_t tangent, tg2, in_direction, normal
            ray_t sp_ray
            complex_t cpx
            double complex n1, n2, n3, cos2, sin2, cos3, sin3, R_p, R_s, T_p, T_s, E1_amp, E2_amp
            double complex n1cos1, n2cos2, n3cos3, n1cos2, n2cos1, n2cos3, n3cos2, ep1, ep2
            double complex M00, M01, M10, M11, phi
            double cosTheta, cos1, P_in
            double tan_mag_sq, c2
            double Two_n1_cos1, aspect
            double wavelength
            int flip
            
        wavelength = self._wavelengths[in_ray.wavelength_idx]
        normal = norm_(orient.normal)
        in_direction = norm_(in_ray.direction)
        sp_ray = convert_to_sp(in_ray[0], normal)
        sp_ray.accumulated_path += sp_ray.length * sp_ray.refractive_index.real
        E1_amp = sp_ray.E1_amp.real + 1.0j*sp_ray.E1_amp.imag
        E2_amp = sp_ray.E2_amp.real + 1.0j*sp_ray.E2_amp.imag
        cosTheta = dotprod_(normal, in_direction)
        cos1 = fabs(cosTheta)
        sin1 = sqrt(1 - cos1*cos1)
        
        n2 = self.n_coating_.real + 1.0j*self.n_coating_.imag
        
        if cosTheta < 0.0: 
            #ray incident from outside going inwards
            n1 = self.n_outside_.real + 1.0j*self.n_outside_.imag
            n3 = self.n_inside_.real + 1.0j*self.n_inside_.imag
            flip = 1
            #print "out to in", n1, n2
        else:
            n1 = self.n_inside_.real + 1.0j*self.n_inside_.imag
            n3 = self.n_outside_.real + 1.0j*self.n_outside_.imag
            flip = -1
            #print "in to out", n1, n2
            
        #apply Snell's law. These become complex.
        sin2 = (n1*sin1/n2)
        cos2 = csqrt(1 - sin2*sin2)
        
        sin3 = (n1*sin1/n3)
        cos3 = csqrt(1 - sin3*sin3)
        
        cosThetaNormal = multvs_(normal, cosTheta)
        
        #incoming power
        P_in =  n1.real*(E1_amp.real**2 + E1_amp.imag**2 + \
                         E2_amp.real**2 + E2_amp.imag**2)
        
        if P_in==0.0:
            return

        #Fresnel equations for reflection and transmission
        n1cos1 = n1*cos1
        n2cos2 = n2*cos2
        n3cos3 = n3*cos3
        dwc = 2*M_PI*self.thickness/wavelength
        phi = -I*dwc*(n2 - sin2*sin2)/cos2
        ep1 = cexp(phi)/(4*n2cos2*n3cos3)
        ep2 = cexp(-2*phi)
        M00 = -ep1*( (n1cos1-n2cos2)*(n2cos2+n3cos3) + 
                     (n1cos1+n2cos2)*(n2cos2-n3cos3)*ep2 )
                     
        M01 = ep1*( (n1cos1-n2cos2)*(n2cos2-n3cos3)*ep2 + 
                     (n1cos1+n2cos2)*(n2cos2+n3cos3) )
        
        M10 = ep1*( (n1cos1-n2cos2)*(n2cos2-n3cos3) + 
                     (n1cos1+n2cos2)*(n2cos2+n3cos3)*ep2 )
        
        M11 = -ep1*( (n1cos1-n2cos2)*(n2cos2+n3cos3)*ep2 + 
                     (n1cos1+n2cos2)*(n2cos2-n3cos3) )
        
        R_s = -M00/M01
        T_s = M10 + M11*R_s
        
        n1cos2 = n1*cos2
        n2cos1 = n2*cos1
        n2cos3 = n2*cos3
        n3cos2 = n3*cos2
        M00 = -ep1*( (n1cos2-n2cos1)*(n2cos3+n3cos2) +
                      (n1cos2+n2cos1)*(n2cos3-n3cos2)*ep2 )
        
        M01 = ep1*( (n1cos2-n2cos1)*(n2cos3-n3cos2)*ep2 +
                      (n1cos2+n2cos1)*(n2cos3+n3cos2) )
        
        M10 = ep1*( (n1cos2-n2cos1)*(n2cos3-n3cos2) +
                      (n1cos2+n2cos1)*(n2cos3+n3cos2)*ep2 )
        
        M11 = -ep1*( (n1cos2-n2cos1)*(n2cos3+n3cos2)*ep2 +
                      (n1cos2+n2cos1)*(n2cos3-n3cos2) )
        
        R_p = -M00/M01
        T_p = M10 + M11*R_p
        
        #modify in place to get reflected amplitudes
        R_s *= E1_amp
        R_p *= E2_amp
        
        
        if ( n1.real*(cabs(R_s)**2 + cabs(R_p)**2)/P_in ) > self.reflection_threshold:
            reflected = subvv_(in_direction, multvs_(cosThetaNormal, 2))
            sp_ray.origin = point
            sp_ray.normal = normal
            sp_ray.direction = reflected
            sp_ray.length = INF
            sp_ray.E1_amp.real = R_s.real
            sp_ray.E1_amp.imag = R_s.imag
            sp_ray.E2_amp.real = -R_p.real
            sp_ray.E2_amp.imag = -R_p.imag
            sp_ray.parent_idx = idx
            sp_ray.refractive_index.real = n1.real
            sp_ray.refractive_index.imag = n1.imag
            sp_ray.ray_type_id |= REFL_RAY
            new_rays.add_ray_c(sp_ray)
            
        #normal transmission            
        tangent = subvv_(in_direction, cosThetaNormal)
        tg2 = multvs_(tangent, n1.real/n3.real) #This is an approximation for complex N
        tan_mag_sq = mag_sq_(tg2)
        c2 = sqrt(1-tan_mag_sq)
        transmitted = subvv_(tg2, multvs_(normal, c2*flip))
        
        aspect = sqrt(cos3.real/cos1)
        
        #modify in place to get transmitted amplitudes
        T_s *= (E1_amp*aspect)
        T_p *= (E2_amp*aspect)
        
        if ( n3.real*(cabs(T_s)**2 + cabs(T_p)**2)/P_in ) > self.transmission_threshold:
            sp_ray.origin = point
            sp_ray.normal = normal
            sp_ray.direction = transmitted
            sp_ray.length = INF
            sp_ray.E1_amp.real = T_s.real
            sp_ray.E1_amp.imag = T_s.imag
            sp_ray.E2_amp.real = T_p.real
            sp_ray.E2_amp.imag = T_p.imag
            sp_ray.parent_idx = idx
            sp_ray.refractive_index.real = n3.real
            sp_ray.refractive_index.imag = n3.imag
            sp_ray.ray_type_id &= ~REFL_RAY
            new_rays.add_ray_c(sp_ray)
    
vacuum = BaseDispersionCurve(0,np.array([1.0,]))
    
            
cdef class CoatedDispersiveMaterial(InterfaceMaterial):
    """
    A full implementation of the Fresnel Equations for a single-layer 
    coated dielectric interface. The refractive index for each ray is obtained
    by look up of the provided dispersio-curve objects for the materials on
    each side of the interface.
    
    :param dispersion_inside: The dispersion curve for the inside side of the interface
    :type dispersion_inside: raypier.core.cmaterials.BaseDispersionCurve
    :param dispersion_outside: The dispersion curve for the "outside" of the interface
    :type dispersion_outside: raypier.core.cmaterials.BaseDispersionCurve
    :param dispersion_coating: The dispersion curve for the coating material
    :type dispersion_coating: raypier.core.cmaterials.BaseDispersionCurve
    :param double coating_thickness: The coating thickness, in microns
    :param double reflection_threshold: Sets the amplitude threshold for generating a reflected ray.
    :param double transmission_threshold: Sets the amplitude threshold for generating a transmitted ray.
    """
    cdef:
        public np_.npy_complex128[:] n_inside, n_outside, n_coating
        public BaseDispersionCurve dispersion_inside
        public BaseDispersionCurve dispersion_outside
        public BaseDispersionCurve dispersion_coating
        public double coating_thickness #in microns
        public double reflection_threshold, transmission_threshold
        
    def __cinit__(self, **kwds):
        self.reflection_threshold = kwds.get('reflection_threshold', 0.1)
        self.transmission_threshold = kwds.get('transmission_threshold', 0.1)
        self.dispersion_inside = kwds.get("dispersion_inside",vacuum)
        self.dispersion_outside = kwds.get("dispersion_outside",vacuum)
        self.dispersion_coating = kwds.get("dispersion_coating",vacuum)
        self.coating_thickness = kwds.get("coating_thickness", 0.1)
        
    cdef on_set_wavelengths(self):
        cdef:
            double[:] wavelengths = self._wavelengths
        self.n_inside = self.dispersion_inside.c_evaluate_n(wavelengths)
        self.n_outside = self.dispersion_outside.c_evaluate_n(wavelengths)
        self.n_coating = self.dispersion_coating.c_evaluate_n(wavelengths)
    
    @cython.cdivision(True)
    cdef void eval_child_ray_c(self,
                            ray_t *in_ray, 
                            unsigned int idx, 
                            vector_t point,
                            orientation_t orient,
                            RayCollection new_rays):
        """
           ray - the ingoing ray
           idx - the index of ray in it's RayCollection
           point - the position of the intersection (in global coords)
           normal - the outward normal vector for the surface
        """
        cdef:
            vector_t cosThetaNormal, reflected, transmitted
            vector_t tangent, tg2, in_direction, normal
            ray_t sp_ray
            complex_t cpx
            double complex n1, n2, n3, cos2, sin2, cos3, sin3, R_p, R_s, T_p, T_s, E1_amp, E2_amp
            double complex n1cos1, n2cos2, n3cos3, n1cos2, n2cos1, n2cos3, n3cos2, ep1, ep2
            double complex M00, M01, M10, M11, phi
            double cosTheta, cos1, P_in
            double tan_mag_sq, c2
            double Two_n1_cos1, aspect
            double wavelength
            int flip
            np_.npy_complex128 ctemp
            
        wavelength = self._wavelengths[in_ray.wavelength_idx]
        normal = norm_(orient.normal)
        in_direction = norm_(in_ray.direction)
        sp_ray = convert_to_sp(in_ray[0], normal)
        sp_ray.accumulated_path += sp_ray.length * sp_ray.refractive_index.real
        
        E1_amp = sp_ray.E1_amp.real + 1.0j*sp_ray.E1_amp.imag
        E2_amp = sp_ray.E2_amp.real + 1.0j*sp_ray.E2_amp.imag
        cosTheta = dotprod_(normal, in_direction)
        cos1 = fabs(cosTheta)
        sin1 = sqrt(1 - cos1*cos1)
        
        ctemp = self.n_coating[in_ray.wavelength_idx]
        n2 = ctemp.real + I*ctemp.imag#self.n_coating_.real + 1.0j*self.n_coating_.imag

        if cosTheta < 0.0: 
            #ray incident from outside going inwards
            ctemp = self.n_outside[in_ray.wavelength_idx]
            n1 = ctemp.real + I*ctemp.imag#self.n_outside_.real + 1.0j*self.n_outside_.imag
            ctemp = self.n_inside[in_ray.wavelength_idx]
            n3 = ctemp.real + I*ctemp.imag#self.n_inside_.real + 1.0j*self.n_inside_.imag
            flip = 1
            #print( "out to in", n1, n3)
        else:
            ctemp = self.n_inside[in_ray.wavelength_idx]
            n1 = ctemp.real + I*ctemp.imag
            ctemp = self.n_outside[in_ray.wavelength_idx]
            n3 = ctemp.real + I*ctemp.imag
            flip = -1
            #print( "in to out", n1, n3 )
            
        #apply Snell's law. These become complex.
        sin2 = (n1*sin1/n2)
        cos2 = csqrt(1 - sin2*sin2)
        
        sin3 = (n1*sin1/n3)
        cos3 = csqrt(1 - sin3*sin3)
        
        cosThetaNormal = multvs_(normal, cosTheta)
        
        #incoming power
        P_in =  n1.real*(E1_amp.real**2 + E1_amp.imag**2 + \
                         E2_amp.real**2 + E2_amp.imag**2)
        
        if P_in==0.0:
            return

        #Fresnel equations for reflection and transmission
        n1cos1 = n1*cos1
        n2cos2 = n2*cos2
        n3cos3 = n3*cos3
        dwc = 2*M_PI*self.coating_thickness/wavelength
        phi = -I*dwc*(n2 - sin2*sin2)/cos2
        ep1 = cexp(phi)/(4*n2cos2*n3cos3)
        ep2 = cexp(-2*phi)
        M00 = -ep1*( (n1cos1-n2cos2)*(n2cos2+n3cos3) + 
                     (n1cos1+n2cos2)*(n2cos2-n3cos3)*ep2 )
                     
        M01 = ep1*( (n1cos1-n2cos2)*(n2cos2-n3cos3)*ep2 + 
                     (n1cos1+n2cos2)*(n2cos2+n3cos3) )
        
        M10 = ep1*( (n1cos1-n2cos2)*(n2cos2-n3cos3) + 
                     (n1cos1+n2cos2)*(n2cos2+n3cos3)*ep2 )
        
        M11 = -ep1*( (n1cos1-n2cos2)*(n2cos2+n3cos3)*ep2 + 
                     (n1cos1+n2cos2)*(n2cos2-n3cos3) )
        
        R_s = -M00/M01
        T_s = M10 + M11*R_s
        
        n1cos2 = n1*cos2
        n2cos1 = n2*cos1
        n2cos3 = n2*cos3
        n3cos2 = n3*cos2
        M00 = -ep1*( (n1cos2-n2cos1)*(n2cos3+n3cos2) +
                      (n1cos2+n2cos1)*(n2cos3-n3cos2)*ep2 )
        
        M01 = ep1*( (n1cos2-n2cos1)*(n2cos3-n3cos2)*ep2 +
                      (n1cos2+n2cos1)*(n2cos3+n3cos2) )
        
        M10 = ep1*( (n1cos2-n2cos1)*(n2cos3-n3cos2) +
                      (n1cos2+n2cos1)*(n2cos3+n3cos2)*ep2 )
        
        M11 = -ep1*( (n1cos2-n2cos1)*(n2cos3+n3cos2)*ep2 +
                      (n1cos2+n2cos1)*(n2cos3-n3cos2) )
        
        R_p = -M00/M01
        T_p = M10 + M11*R_p
        
        #modify in place to get reflected amplitudes
        R_s *= E1_amp
        R_p *= E2_amp
        
        
        if ( n1.real*(cabs(R_s)**2 + cabs(R_p)**2)/P_in ) > self.reflection_threshold:
            reflected = subvv_(in_direction, multvs_(cosThetaNormal, 2))
            sp_ray.origin = point
            sp_ray.normal = normal
            sp_ray.direction = reflected
            sp_ray.length = INF
            sp_ray.E1_amp.real = R_s.real
            sp_ray.E1_amp.imag = R_s.imag
            sp_ray.E2_amp.real = -R_p.real
            sp_ray.E2_amp.imag = -R_p.imag
            sp_ray.parent_idx = idx
            sp_ray.refractive_index.real = n1.real
            sp_ray.refractive_index.imag = n1.imag
            sp_ray.ray_type_id |= REFL_RAY
            new_rays.add_ray_c(sp_ray)
            
        #normal transmission            
        tangent = subvv_(in_direction, cosThetaNormal)
        tg2 = multvs_(tangent, n1.real/n3.real) #This is an approximation for complex N
        tan_mag_sq = mag_sq_(tg2)
        c2 = sqrt(1-tan_mag_sq)
        transmitted = subvv_(tg2, multvs_(normal, c2*flip))
        
        aspect = sqrt(cos3.real/cos1)
        
        #modify in place to get transmitted amplitudes
        T_s *= (E1_amp*aspect)
        T_p *= (E2_amp*aspect)
        
        if ( n3.real*(cabs(T_s)**2 + cabs(T_p)**2)/P_in ) > self.transmission_threshold:
            sp_ray.origin = point
            sp_ray.normal = normal
            sp_ray.direction = transmitted
            sp_ray.length = INF
            sp_ray.E1_amp.real = T_s.real
            sp_ray.E1_amp.imag = T_s.imag
            sp_ray.E2_amp.real = T_p.real
            sp_ray.E2_amp.imag = T_p.imag
            sp_ray.parent_idx = idx
            sp_ray.refractive_index.real = n3.real
            sp_ray.refractive_index.imag = n3.imag
            sp_ray.ray_type_id &= ~REFL_RAY
            new_rays.add_ray_c(sp_ray)
            
    cdef para_t eval_parabasal_ray_c(self,
                                     ray_t *base_ray, 
                                     vector_t direction, #incoming ray direction
                                   vector_t point, #position of intercept
                                   orientation_t orient,
                                   unsigned int ray_type_id, #bool, if True, it's a reflected ray
                                   ):
        cdef:
            vector_t cosThetaNormal, normal, out_dir
            para_t para_out
            double cosTheta, n1, n2
            int flip
        
        normal = norm_(orient.normal)
        direction = norm_(direction)
        cosTheta = dotprod_(normal, direction)
        cosThetaNormal = multvs_(normal, cosTheta)
        
        if cosTheta < 0.0: 
            #ray incident from outside going inwards
            n1 = self.n_outside[base_ray.wavelength_idx].real#self.n_outside_.real
            n2 = self.n_inside[base_ray.wavelength_idx].real
            flip = 1
        else:
            n1 = self.n_inside[base_ray.wavelength_idx].real
            n2 = self.n_outside[base_ray.wavelength_idx].real
            flip = -1
        
        if ray_type_id & REFL_RAY:
            out_dir = subvv_(direction, multvs_(cosThetaNormal, 2))
        else:
            tangent = subvv_(direction, cosThetaNormal)
            tg2 = multvs_(tangent, n1/n2)
            tan_mag_sq = mag_sq_(tg2)
            c2 = sqrt(1-tan_mag_sq)
            out_dir = subvv_(tg2, multvs_(normal, c2*flip))
        
        para_out.direction = out_dir 
        para_out.origin = point
        para_out.normal = normal
        para_out.length = INF
        return para_out
            

cdef class DiffractionGratingMaterial(InterfaceMaterial):
    """
    A specialised material modelling the behaviour of a reflective diffraction grating.
    
    :param double lines_per_mm: The grating line density
    :param int order: The order-of-diffraction
    :param double efficiency: A value from 0.0 to 1.0 giving the reflection efficiency of the grating
    :param origin: A vector (3-tuple) indicating the origin of the grating. Unimportant in most cases. 
                    This affects the "grating phase" which is unobservable in most situations.
    :type origin: (double, double, double)
    
    """
    ###A perfect reflection diffraction grating
    cdef:
        public double lines_per_mm #: The line-density of the grating
        public int order #order of diffraction
        public double efficiency
        vector_t origin_
        
    def __cinit__(self, **kwds):
        self.lines_per_mm = kwds.get("lines_per_mm", 1000)
        self.order = kwds.get("order", 1)
        self.efficiency = kwds.get("efficiency", 1.0)
        self.origin = kwds.get("origin", (0.0,0.0,0.0))
        
    property origin: 
        def __get__(self):
            cdef vector_t o = self.origin_
            return (o.x, o.y, o.z)
        
        def __set__(self, o):
            self.origin_.x = o[0]
            self.origin_.y = o[1]
            self.origin_.z = o[2]
    
    cdef void eval_child_ray_c(self,
                            ray_t *in_ray, 
                            unsigned int idx, 
                            vector_t point,
                            orientation_t orient,
                            RayCollection new_rays):
        """
           ray - the ingoing ray
           idx - the index of ray in it's RayCollection
           point - the position of the intersection (in global coords)
           normal - the outward normal vector for the surface
        """
        cdef:
            vector_t reflected, normal, tangent, tangent2
            ray_t sp_ray
            complex_t n_ray
            double k_x, k_y, k_z
            double wavelen, line_spacing
            int sign
        
        #Create surface unit axes
        normal = norm_(orient.normal)
        tangent = norm_(orient.tangent) #this is the grating wavevector direction
        tangent2 = cross_(normal, tangent) #parallel to grating lines
        
        wavelen = self._wavelengths[in_ray.wavelength_idx]
        line_spacing = 1000.0 / self.lines_per_mm #in microns
        
        reflected = norm_(in_ray.direction) #ensure unit vector
        #Break incoming direction vector into components along surface unit axes
        k_z = dotprod_(normal, reflected)
        k_y = dotprod_(tangent2, reflected)
        k_x = dotprod_(tangent, reflected)
        
        if k_z<0.0:#invert for a reflecting grating
            sign = 1
        else:
            sign = -1
        
        n_ray = in_ray[0].refractive_index
        k_x = k_x - self.order*wavelen/(line_spacing*n_ray.real)
        #y-component (parallel to grating lines) doesn't change
        
        k_z = 1 - (k_x*k_x) - (k_y*k_y) #Apply Pythagoras to get z-component
        if k_z < 0: #diffracted ray is evanescent
            return
        k_z = sign * sqrt(k_z) 
        
        #Add the x- , y- and z-components together
        reflected = multvs_(tangent, k_x)
        reflected = addvv_(reflected, multvs_(tangent2, k_y))
        reflected = addvv_(reflected, multvs_(normal, k_z))
                
        sp_ray = convert_to_sp(in_ray[0], normal)
        sp_ray.accumulated_path += sp_ray.length * sp_ray.refractive_index.real
        
        sp_ray.origin = point
        sp_ray.normal = normal
        sp_ray.direction = reflected
        sp_ray.E1_amp.real = -sp_ray.E1_amp.real * self.efficiency
        sp_ray.E1_amp.imag = -sp_ray.E1_amp.imag * self.efficiency
        sp_ray.E2_amp.real = sp_ray.E2_amp.real * self.efficiency
        sp_ray.E2_amp.imag = sp_ray.E2_amp.imag * self.efficiency
        sp_ray.parent_idx = idx
        sp_ray.ray_type_id |= REFL_RAY
        
        ### This is the mysterious Grating Phase
        #Need to convert origin offset to microns since line-spacing is in microns
        sp_ray.phase += 1000.0*dotprod_(subvv_(self.origin_, point), tangent)*self.order*2*M_PI/line_spacing
        
        new_rays.add_ray_c(sp_ray)
        
    cdef para_t eval_parabasal_ray_c(self,
                                     ray_t *base_ray, 
                                     vector_t direction, #incoming ray direction
                                   vector_t point, #position of intercept
                                   orientation_t orient,
                                   unsigned int ray_type_id, #bool, if True, it's a reflected ray
                                   ):        
        cdef:
            para_t para_out
            vector_t reflected, normal, tangent, tangent2
            ray_t sp_ray
            complex_t n_ray
            double k_x, k_y, k_z
            double wavelen, line_spacing
            int sign
        
        #Create surface unit axes
        normal = norm_(orient.normal)
        tangent = norm_(orient.tangent) #this is the grating wavevector direction
        tangent2 = cross_(normal, tangent) #parallel to grating lines
        
        wavelen = self._wavelengths[base_ray.wavelength_idx]
        line_spacing = 1000.0 / self.lines_per_mm #in microns
        
        reflected = norm_(direction) #ensure unit vector
        #Break incoming direction vector into components along surface unit axes
        k_z = dotprod_(normal, reflected)
        k_y = dotprod_(tangent2, reflected)
        k_x = dotprod_(tangent, reflected)
        
        if k_z<0.0:#invert for a reflecting grating
            sign = 1
        else:
            sign = -1
        
        n_ray = base_ray.refractive_index #Should this be the parent refractive index?
        k_x = k_x - self.order*wavelen/(line_spacing*n_ray.real)
        #y-component (parallel to grating lines) doesn't change
        
        k_z = 1 - (k_x*k_x) - (k_y*k_y) #Apply Pythagoras to get z-component
        if k_z < 0: #diffracted ray is evanescent
            print("Error. Parabasal reflection is imaginary!")
            #return #Then we're in trouble!
        k_z = sign * sqrt(k_z) 
        
        #Add the x- , y- and z-components together
        reflected = multvs_(tangent, k_x)
        reflected = addvv_(reflected, multvs_(tangent2, k_y))
        reflected = addvv_(reflected, multvs_(normal, k_z))
                
        para_out.direction = reflected
        para_out.origin = point
        para_out.normal = normal
        para_out.length=INF        
        
        return para_out 
        
        
cdef class CircularApertureMaterial(InterfaceMaterial):
    """Similar to the TransparentMaterial i.e. it generates an
    outgoing ray with identical direction, polarisation etc.
    to the incoming ray. The material attenuates the E_field amplitudes
    according to the radial distance from the surface origin. 
    
    :param double outer_radius: Rays passing outside this radius are not intercepted.
    :param double radius: The radius of the inner hole.
    :param double edge_width: Rays passing through the inner hole are attenuated according to their 
                    proximity to the edge. The edge-width sets the width of the error-function (errf)
                    used to calculate the attenuation.
    :param int invert: Inverts the aperture to make a field-stop.
    :param origin: The centre point of the aperture.
    :type origin: (double, double, double)
    """
    cdef:
        public double outer_radius
        public double radius
        public double edge_width
        public int invert
        vector_t origin_
        
    def __cinit__(self, **kwds):
        self.outer_radius = kwds.get("outer_radius", 25.0)
        self.radius = kwds.get("radius", 15.0)
        self.edge_width = kwds.get("edge_width", 1.0)
        self.origin = kwds.get("origin", (0.0,0.0,0.0))
        self.invert = kwds.get("invert", 0)
        
    property origin: 
        def __get__(self):
            cdef vector_t o = self.origin_
            return (o.x, o.y, o.z)
        
        def __set__(self, o):
            self.origin_.x = o[0]
            self.origin_.y = o[1]
            self.origin_.z = o[2]
        
    cdef void eval_child_ray_c(self,
                            ray_t *in_ray, 
                            unsigned int idx, 
                            vector_t point,
                            orientation_t orient,
                            RayCollection new_rays):
        cdef:
            vector_t normal
            ray_t sp_ray
            double atten, r
            double width = self.edge_width
            
        r = sqrt(mag_sq_(subvv_(self.origin_, point)))
        if (r > self.outer_radius):
            return
        atten = 0.5 + 0.5*erf((self.radius - r)/width)
        if self.invert:
            atten = 1 - atten
        
        normal = norm_(orient.normal)
        sp_ray = convert_to_sp(in_ray[0], normal)
        sp_ray.accumulated_path += sp_ray.length * sp_ray.refractive_index.real

        sp_ray.origin = point
        sp_ray.normal = normal
        sp_ray.parent_idx = idx
        sp_ray.ray_type_id &= ~REFL_RAY
        
        sp_ray.E1_amp.real *= atten
        sp_ray.E1_amp.imag *= atten
        sp_ray.E2_amp.real *= atten
        sp_ray.E2_amp.imag *= atten
        
        new_rays.add_ray_c(sp_ray)


cdef class RectangularApertureMaterial(InterfaceMaterial):
    """
    A rectangular aperture.
    
    :param double outer_width:
    :param double outer_height:
    :param double width:
    :param double height:
    :param double edge_width:
    :param int invert:
    :param origin:
    :type origin: (double, double, double)
    """
    cdef:
        public double outer_width
        public double outer_height
        public double width
        public double height
        public double edge_width
        public int invert
        vector_t origin_
        
    def __cinit__(self, **kwds):
        self.outer_width = kwds.get("outer_width", 15.0)
        self.outer_height = kwds.get("outer_height", 20.0)
        self.width = kwds.get("width", 5.0)
        self.height = kwds.get("height", 10.0)
        self.edge_width = kwds.get("edge_width", 1.0)
        self.origin = kwds.get("origin", (0.0,0.0,0.0))
        self.invert = kwds.get("invert", 0)
        
    property origin: 
        def __get__(self):
            cdef vector_t o = self.origin_
            return (o.x, o.y, o.z)
        
        def __set__(self, o):
            self.origin_.x = o[0]
            self.origin_.y = o[1]
            self.origin_.z = o[2]
        
    cdef void eval_child_ray_c(self,
                            ray_t *in_ray, 
                            unsigned int idx, 
                            vector_t point,
                            orientation_t orient,
                            RayCollection new_rays):
        cdef:
            vector_t normal, p
            ray_t sp_ray
            double atten, px, py
            double width = self.edge_width
            double x = self.width/2.
            double y = self.height/2.
            
        p = subvv_(point, self.origin_)
        px = dotprod_(p,orient.tangent)
        py = dotprod_(p,cross_(orient.normal,orient.tangent))
            
        if fabs(px) > self.outer_width/2.:
            return 
        if fabs(py) > self.outer_height/2.:
            return 
            
        atten = 0.5 - 0.5*erf((px - x)/width)
        atten *= 0.5 - 0.5*erf(-(px + x)/width)
        atten *= 0.5 - 0.5*erf((py - y)/width)
        atten *= 0.5 - 0.5*erf(-(py + y)/width)
        
        if self.invert:
            atten = 1 - atten
        
        normal = norm_(orient.normal)
        sp_ray = convert_to_sp(in_ray[0], normal)
        sp_ray.accumulated_path += sp_ray.length * sp_ray.refractive_index.real

        sp_ray.origin = point
        sp_ray.normal = normal
        sp_ray.parent_idx = idx
        sp_ray.ray_type_id &= ~REFL_RAY
        
        sp_ray.E1_amp.real *= atten
        sp_ray.E1_amp.imag *= atten
        sp_ray.E2_amp.real *= atten
        sp_ray.E2_amp.imag *= atten
        
        new_rays.add_ray_c(sp_ray)
        
        
cdef class ResampleGaussletMaterial(InterfaceMaterial):
    """
    This is a special pseudo-material which generates new rays not by
    the normal process of refraction or reflection of an incoming ray,
    but by computing a new set of Gausslets by computing the E-field
    at a set of grid points and launching new Gausslets from these points.
    
    The material needs the set of new launch-positions to be given up front (i.e. before tracing). 
    
    :param int size: The size of the new GaussletCollection to allocate up front.
    :param callable eval_func: A callable taking a GaussletCollection as its single argument.
            The new outgoing rays will be returned by this callable.
    """
    cdef:
        public int capture_count
        public GaussletCollection captured_rays
        object _evaluation_func
        
    def __cinit__(self, **kwds):
        size = kwds.get("size", 2)
        func = kwds.get("eval_func", None)
        self.captured_rays = GaussletCollection(size)
        if func is not None:
            self.eval_func = func
        
    property eval_func:
        def __get__(self):
            return self._evaluation_func
        
        def __set__(self, obj):
            if not callable(obj):
                raise ValueError("The eval_func property must be a callable.")
            self._evaluation_func = obj
            
    def is_decomp_material(self):
        return True
    
    cdef void eval_child_ray_c(self,
                            ray_t *in_ray, 
                            unsigned int idx, 
                            vector_t point,
                            orientation_t orient,
                            RayCollection new_rays):
        """
        Capture each incident ray in an internal GaussletCollection.
        """
        cdef:
            gausslet_t *in_g
            
        if in_ray[0].ray_type_id & 2: #check GAUSSLET bit
            in_g = <gausslet_t *>in_ray #oooh, dangerous! Works because each gausslet_t starts with a ray_t
            self.captured_rays.add_gausslet_c(in_g[0])
            self.capture_count += 1
    
    cdef void eval_decomposed_rays_c(self, GaussletCollection new_rays):
        """
        Here we compute the new rays and append them to the new_rays object.
        """
        cdef:
            GaussletCollection gc
            
        self.captured_rays._wavelengths = new_rays._wavelengths
        gc = self._evaluation_func(self.captured_rays)
        new_rays.extend_c(gc)
        #Clear the captured rays
        self.captured_rays.n_rays = 0




cdef extern from "math.h":
    double M_PI
    double sqrt(double) nogil
    double atan2 (double y, double x )
    double pow(double x, double y)
    double fabs(double)
    double cos(double)
    double sin(double)
    double acos(double)
    

IF UNAME_SYSNAME == "Windows":
    cdef extern from "complex.h":
        double complex csqrt "sqrt" (double complex) nogil
        double cabs "abs" (double complex) nogil
        double complex cexp "exp" (double complex) nogil
    
    cdef double complex I = 1j
ELSE:
    cdef extern from "complex.h":
        double complex csqrt (double complex) nogil
        double cabs (double complex) nogil
        double complex cexp (double complex) nogil
        double complex I    
        
from cython.parallel import prange


from .ctracer cimport Face, sep_, \
        vector_t, ray_t, FaceList, subvv_, dotprod_, mag_sq_, norm_,\
            addvv_, multvs_, mag_, transform_t, Transform, transform_c,\
                rotate_c, cross_, RayCollection, complex_t, \
                GaussletCollection, gausslet_t, para_t

from .ctracer import ray_dtype
cimport cython
import numpy as np
cimport numpy as np_


cpdef check_this(ray_t[:] rays):
    return rays.shape

cdef:
    double complex rootI=csqrt(I)


@cython.boundscheck(False)  # Deactivate bounds checking
@cython.wraparound(False)   # Deactivate negative indexing.
@cython.cdivision(True)
cpdef  sum_gaussian_modes(RayCollection rays,
                          double complex[:,:] modes, 
                          np_.npy_float64[:] wavelengths,
                          np_.npy_float64[:,:] points):
    """
    Compute the E-field at the given points by propagation of the
    given rays, with phase and mode-coefficients.
    """
    cdef:
        size_t iray, ipt 
        size_t Nray=rays.n_rays
        size_t Npt=points.shape[0]
        
        ray_t ray
        complex_t[:,:] out = np.zeros((Npt,3), dtype=np.complex128)
        vector_t pt, H, E
        double complex U, A, B, C, E1, E2, detG0, kz
        double x,y,z, phase, k, inv_root_area, invk
    
    with nogil:
        for iray in range(Nray):
            ray = rays.rays[iray]
            E = norm_(ray.E_vector)
            H = norm_(cross_(ray.direction, E))
            k = 2000.0*M_PI/wavelengths[ray.wavelength_idx]
            ### The accumulated path length already includes the refractive index of up-stream rays
            ### Hence, need to calculate it before applying the refractive index of the last leg.
            phase = ray.phase + (ray.accumulated_path*k)
            
            #kz = ray.refractive_index.real + I*ray.refractive_index.imag
            kz = ray.refractive_index
            kz *= k
            invk = 2./kz.real
            A = modes[iray, 0]
            B = modes[iray, 1]
            C = modes[iray, 2]
            
            ###normalisation factor 1/root(area). Yes, there really is a double square root.
            inv_root_area = sqrt(sqrt(A.imag*C.imag -(B.imag*B.imag))*(2.0/M_PI))
            
            A.imag *= invk
            B.imag *= invk
            C.imag *= invk
            detG0 = (A*C) - (B*B)
            
            for ipt in prange(Npt):
                pt.x = points[ipt,0]
                pt.y = points[ipt,1]
                pt.z = points[ipt,2]
                pt = subvv_(pt, ray.origin)
                
                U = calc_mode_U(A,B,C, detG0, pt, E, H, ray.direction, kz, phase, inv_root_area)
                
                E1 = ray.E1_amp * U
                E2 = ray.E2_amp * U
                
                out[ipt,0] += (E1*E.x + E2*H.x)
                out[ipt,1] += (E1*E.y + E2*H.y)
                out[ipt,2] += (E1*E.z + E2*H.z)
    return np.asarray(out)

        
cdef double complex calc_mode_U(double complex A,
                                double complex B,
                                double complex C,
                                double complex detG0,
                                vector_t pt, 
                                vector_t E, 
                                vector_t H,
                                vector_t direction,
                                double complex k,
                                double phase, 
                                double inv_root_area) nogil:
    cdef:
        double complex denom, AA, CC, U
        double x,y,z
        
    x = dotprod_(pt, E)
    y = dotprod_(pt, H)
    z = dotprod_(pt, direction)
    ### Where did this factor of 2 come from?
    denom = (1 + (z*(A+C)) + (z*z)*detG0) *2
    AA = (A + z*detG0)/denom
    CC = (C + z*detG0)/denom
    U = cexp( I*(phase + k*(z + AA*(x*x) + B*(2*x*y)/denom + CC*(y*y) ) ) )
    ###Normalisation factor
    ### I've added an extra 1I factor here in order to rotate the phase away from the 0/2pi wrapping point.
    ### Seems like the csqrt (or cpow) functions struggle with accuracy near phase zero where we're summing 
    ### many values straddling the wrapping point
    U /= csqrt(((1 + z*A)*(1 + z*C) - (z*B)*(z*B))*I)
    
    ###normalise by ray initial area
    ### multiply by sqrt(I) to reverse the effect of the I factor applied above
    U *= inv_root_area*rootI
    
    return U


@cython.boundscheck(False)
@cython.wraparound(False)
cdef void evaluate_one_mode(np_.npy_complex128[:] out, double[:] x, double[:] y, 
                            double[:] dx, double[:] dy, double blending, int size) nogil:
    ### Brute force evaluation of the linear least squares fit to the x,y,dx,dy coefs
    ### We solve A.T*A x = A.T b which (by magical means) gives the least squares fit to A x = b
    ### A.T*A and A.T b have 3 rows i.e. we just need to solve 3 equations for 3 unknowns.
    ### Even better, the matrices are symmetric. Solving in sympy gives the closed form of the solution
    ### as fairly simple expressions 
    cdef:
        double a00=0, a01=0, a02=0, a11=0, a12=0, a22=0, b0=0, b1=0, b2=0
        int i
        double xi2, yi2
    
    ### Imag parts ###    
    for i in range(size):
        xi2 = x[i]*x[i]
        yi2 = y[i]*y[i]
        a00 += xi2*xi2
        a01 += 2*xi2*x[i]*y[i]
        a02 += xi2*yi2
        a11 += 4*xi2*yi2
        a12 += 2*x[i]*y[i]*yi2
        a22 += yi2*yi2
        b0 += xi2
        b1 += 2*x[i]*y[i]
        b2 += yi2
    
    out[0].imag = blending *(-b0*(a11*a22 - a12**2) + b1*(a01*a22 - a02*a12) - b2*(a01*a12 - a02*a11))/(-a00*a11*a22 + a00*a12**2 + a01**2*a22 - 2*a01*a02*a12 + a02**2*a11)
    out[1].imag = -blending *(b0*(a01*a22 - a02*a12) - b1*(a00*a22 - a02**2) + b2*(a00*a12 - a01*a02))/(-a00*a11*a22 + a00*a12**2 + a01**2*a22 - 2*a01*a02*a12 + a02**2*a11)
    out[2].imag = blending *(-b0*(a01*a12 - a02*a11) + b1*(a00*a12 - a01*a02) - b2*(a00*a11 - a01**2))/(-a00*a11*a22 + a00*a12**2 + a01**2*a22 - 2*a01*a02*a12 + a02**2*a11)
    
    ### Real parts ###
    a00=0
    a01=0
    a02=0
    a11=0
    a12=0
    a22=0
    b0=0
    b1=0
    b2=0 
    for i in range(size):
        xi2 = x[i]*x[i]
        yi2 = y[i]*y[i]
        
        a00 += xi2
        a01 += x[i]*y[i]
        a11 += xi2 + yi2
        a22 += yi2
        
        b0 += dx[i]*x[i]
        b1 += dx[i]*y[i] + dy[i]*x[i]
        b2 += dy[i]*y[i]
        
    a12 = a01*a01 #reuse a12 as an optimisation

    out[0].real = (-a12*b2 + a01*a22*b1 + b0*(a12 - a11*a22))/(a00*a12 - a00*a11*a22 + a12*a22)
    out[1].real = -(a00*a01*b2 - a00*a22*b1 + a01*a22*b0)/(a00*a12 - a00*a11*a22 + a12*a22)
    out[2].real = (a00*a01*b1 - a12*b0 - b2*(a00*a11 - a12))/(a00*a12 - a00*a11*a22 + a12*a22)
    
    return
    
    
def evaluate_modes(double[:,:] x, double[:,:] y, double[:,:] dx, double[:,:] dy, double blending=1.0):
    cdef:
        int n_rays=x.shape[0], row_size=x.shape[1]
        int i
        np_.npy_complex128[:,:] out = np.zeros((n_rays,3), dtype=np.complex128)        
    
    with nogil:
        for i in prange(n_rays):
            evaluate_one_mode(out[i,:], x[i], y[i], dx[i], dy[i], blending, row_size)
        
    return np.asarray(out)

@cython.boundscheck(False)  # Deactivate bounds checking
@cython.wraparound(False)   # Deactivate negative indexing.
def calc_mode_curvature(double[:] rx, double[:] ry, double[:] dx, double[:] dy, 
                        double[:] dx2, double[:] dy2, double[:] dxdy):
    """
    Compute the Gaussian mode curvature values A,B,C at each (rx,ry) point.
    
    Inputs all have length N.
    
    Returns - (A, B, C, x', y', z')
                A,B,C are 1d arrays with the curvature coefficients. 
                x', y' and z' are (N,3) arrays giving the local basis vectors for the modes.
    """
    cdef:
        unsigned int N=len(rx), i
        
        np_.ndarray _A = np.empty((N), 'd')
        np_.ndarray _B = np.empty((N), 'd')
        np_.ndarray _C = np.empty((N), 'd')
        np_.ndarray _x = np.empty((N,3), 'd')
        np_.ndarray _y = np.empty((N,3), 'd')
        np_.ndarray _z = np.empty((N,3), 'd')
        
        double[:] A=_A,B=_B,C=_C
        double[:,:] x=_x,y=_y,z=_z
        
        vector_t a,b,c
        
    for i in range(N):
        #z' vector
        c.x = -dx[i]
        c.y = -dy[i]
        c.z = 1.0
        c = norm_(c)
        
        b.x = 0.0
        b.y = 0.0
        b.z = 1.0
        
        a = norm_(cross_(b,c))
        b = cross_(c,a)
        
        x[i,0] = a.x
        x[i,1] = a.y
        x[i,2] = a.z
        y[i,0] = b.x
        y[i,1] = b.y
        y[i,2] = b.z
        z[i,0] = c.x
        z[i,1] = c.y
        z[i,2] = c.z
        
        Fz = c.x*dx[i] + c.y*dy[i] + c.z
        
        ### This is d2F/dx^2
        A[i] = -( a.x*a.x*dx2[i] + 2*a.x*a.y*dxdy[i] + a.y*a.y*dy2[i] )/Fz
        ### This is d2F/dxdy
        B[i] = -( a.x*b.x*dx2[i] + (a.x*b.y + b.x*a.y)*dxdy[i] + a.y*b.y*dy2[i])/Fz
        ### This is d2F/dy^2
        C[i] = -( b.x*b.x*dx2[i] + 2*b.x*b.y*dxdy[i] + b.y*b.y*dy2[i] )/Fz
        
    return (_A, _B, _C, _x, _y, _z)
    
@cython.boundscheck(False)  # Deactivate bounds checking
@cython.wraparound(False)   # Deactivate negative indexing.
def build_interaction_matrix(double[:] rx, double[:] ry,
                             double[:] tx, double[:] ty,
                             double[:] A, double[:] B, double[:] C, 
                             double[:,:] x, double[:,:] y, double[:,:] z,
                             double wavelength, double spacing, 
                             double max_spacing, double blending,
                             double test_spacing):
    """
    Calculate the sparse matrix (COO format) to represent the complex amplitude of the field E_ij
    at mode i due to the mode j. E_ij = 1 where i==j.
    
    We calculate this in the local coordinate system of the decomposition plane.
    
    
    
    Parameters
    ----------
    
    rx - double[:] array of x-axis coordinates for the ray origins
    ry - double[:] array of y-axis coordinates for the ray origins
    A - double[:] wavefront curvature d2z'/dx'2 in ray-local basis
    B - double[:] wavefront curvature d2z'/dx'dy' in ray-local basis
    C - double[:] wavefront curvature d2z'/dy'2 in ray-local basis
    x,y,z - double[:,3] arrays giving the basis vectors for each ray (z is ray direction vector)
    wavelength - wavelength in microns
    spacing - the spacing between adjacent rays, in mm
    max_spacing - the maximum spacing between rays to calculate the cross-interaction. Controls the sparsity 
                    of the final result
    blending - adjusts the widths of each mode, relative to the spacing. width = spacing/blending
    test_spacing - the spacing used by the test-points
    
    Returns
    -------
    The returned tuple of arrays (data[:], (i[;],j[:]) is suitable to pass to scipy.sparse.coo_matrix().
    """
    
    cdef:
        unsigned int M, N=len(rx), L=len(tx), i, j, ct=0
        vector_t a,b,c,o,pt
        double complex[:] data
        unsigned long[:] xi, xj
        double phase=0.0, k, inv_root_area, impart
        double complex _A, _B, _C, detG0
        
    ##Estimate the size of the output array
    M = N*( 1 + int((7 * max_spacing * max_spacing)/(test_spacing*test_spacing)) ) 
    
    data = np.zeros(M, np.complex128)
    xi = np.zeros(M, np.uint)
    xj = np.zeros(M, np.uint)
    
    k = 2000.0*M_PI/wavelength
    
    impart = 2*(blending*blending)/(k*spacing*spacing)
    
    ###normalisation factor 1/root(area). Yes, there really is a double square root.
    inv_root_area = 1.0 #All modes have the same width
        
    for i in range(N):
        ### Origin of i'th mode
        o.x = rx[i]
        o.y = ry[i]
        o.z = 0.0
        
        a.x = x[i,0]
        a.y = x[i,1]
        a.z = x[i,2]
        b.x = y[i,0]
        b.y = y[i,1]
        b.z = y[i,2]
        c.x = z[i,0]
        c.y = z[i,1]
        c.z = z[i,2]
        
        ### Real parts are curvature, imaginary parts are 1/e^2 widths
        _A.real = A[i]
        _B.real = B[i]
        _C.real = C[i]
        
        _A.imag = impart
        _B.imag = 0 #It's a symmetric mode
        _C.imag = impart 
        
        detG0 = (_A*_C) #- (B*B)
        
        for j in range(L):
            pt.x = tx[j]
            pt.y = ty[j]
            pt.z = 0.0
            pt = subvv_(pt, o)
            
            if ((pt.x*pt.x) + (pt.y*pt.y)) <= (max_spacing*max_spacing):
                xi[ct] = i
                xj[ct] = j
                #if i == j:
                #    data[ct] = 1.0
                #else:
                data[ct] = calc_mode_U(_A, _B, _C, detG0, pt, a, b, c, k, phase, inv_root_area)
                ct += 1
    
    return (np.asarray(data)[:ct], (np.asarray(xi)[:ct], np.asarray(xj)[:ct]))


def apply_mode_curvature(GaussletCollection gc, double[:] A, double[:] B, double[:] C):
    """
    Adds the given wavefront curvature coefficients (i.e. 2nd derivatives) to the 
    given GaussletCollection object.
    """
    cdef:
        unsigned int i,j
        gausslet_t *g
        para_t *para
        ray_t *ray
        vector_t x,y,z, shift, r
        double px, py
        
    for i in range(gc.n_rays):
        g = &(gc.rays[i])
        ray = &(g.base_ray)
        z = ray.direction
        x = ray.E_vector
        y = cross_(x,z)
        for j in range(6):
            para = &(g.para[j])
            r = subvv_(para.origin, ray.origin)
            px = dotprod_(r, x)
            py = dotprod_(r, y)
            shift = multvs_(x, (-A[i]*px - B[i]*py))
            shift = addvv_(shift, multvs_(y, (-C[i]*py - B[i]*px)) )
            para.direction = norm_(addvv_(shift, para.direction))
            ###Not finished
    return 



from fontTools.cffLib import width

cdef extern from "math.h":
    double M_PI
    double sqrt(double) nogil
    double atan2 (double y, double x )
    double pow(double x, double y)
    double fabs(double)
    double cos(double)
    double sin(double)
    double acos(double)


from .ctracer cimport Shape, sep_, \
        vector_t, subvv_, dotprod_, mag_sq_, norm_,\
            addvv_, multvs_, mag_, cross_
            
import numpy as np
            

cdef class LogicalOpShape(Shape):
    def __and__(self, Shape other):
        return BooleanAND(self, other)
    
    def __or__(self, Shape other):
        return BooleanOR(self, other)
    
    def __xor__(self, Shape other):
        return BooleanXOR(self, other)
    
    def __invert__(self):
        return InvertShape(self)
            
            
cdef class InvertShape(LogicalOpShape):
    cdef:
        public Shape shape
        
    def __cinit__(self, Shape shape):
        self.shape = shape
        
    cdef bint point_inside_c(self, double x, double y):
        return 1 & (~self.shape.point_inside_c(x,y))
            
            
cdef class BooleanShape(LogicalOpShape):
    cdef:
        public Shape shape1
        public Shape shape2
        
    def __cinit__(self, Shape shape1, Shape shape2):
        self.shape1 = shape1
        self.shape2 = shape2
        
        
cdef class BooleanAND(BooleanShape):
    cdef bint point_inside_c(self, double x, double y):
        return (<Shape>self.shape1).point_inside_c(x,y) & (<Shape>self.shape2).point_inside_c(x,y)
    
    
cdef class BooleanOR(BooleanShape):
    cdef bint point_inside_c(self, double x, double y):
        return (<Shape>self.shape1).point_inside_c(x,y) | (<Shape>self.shape2).point_inside_c(x,y)
    
    
cdef class BooleanXOR(BooleanShape):
    cdef bint point_inside_c(self, double x, double y):
        return (<Shape>self.shape1).point_inside_c(x,y) ^ (<Shape>self.shape2).point_inside_c(x,y)
    
    
cdef class BasicShape(LogicalOpShape):
    cdef:
        double centre_x, centre_y
        
    def __cinit__(self, **kwds):
        if "centre" in kwds:
            self.centre = kwds["centre"]
        else:
            self.centre_x = kwds.get("centre_x", 0.0)
            self.centre_y = kwds.get("centre_y", 0.0)
        
    property centre:
        def __get__(self):
            return (self.centre_x, self.centre_y)
        
        def __set__(self, v):
            self.centre_x = v[0]
            self.centre_y = v[1]
            
    def __and__(self, Shape other):
        return BooleanAND(self, other)
    
    def __or__(self, Shape other):
        return BooleanOR(self, other)
    
    def __xor__(self, Shape other):
        return BooleanXOR(self, other)
    
    def __invert__(self):
        return InvertShape(self)
                
            
cdef class CircleShape(BasicShape):
    cdef:
        public double radius
        
    def __cinit__(self, **kwds):
        self.radius = kwds.get("radius", 1.0)
        
    cdef bint point_inside_c(self, double x, double y):
        cdef:
            double dx = x-self.centre_x
            double dy = y-self.centre_y
        if (dx*dx) + (dy*dy) < (self.radius*self.radius):
            return 1
        else:
            return 0
        
        
cdef class RectangleShape(BasicShape):
    cdef:
        public double width
        public double height
        
    def __cinit__(self, **kwds):
        self.width = kwds.get("width", 5.0)
        self.height = kwds.get("height", 7.0)
        
    cdef bint point_inside_c(self, double x, double y):
        cdef:
            double dx = x-self.centre_x
            double dy = y-self.centre_y
            
        if (2*fabs(dx) < self.width) and (2*fabs(dy)) < self.height:
            return 1
        else:
            return 0


cdef class PolygonShape(BasicShape):
    cdef:
        double[:,:] _coordinates
        
    property coordinates:
        def __get__(self):
            return np.asarray(self._coordinates)
        
        def __set__(self, val):
            self._coordinates = val
            
    cdef bint point_inside_c(self, double X, double Y):
        cdef:
            int i, size, ct=0
            double y1, y2, x1, x2
            double[:,:] pts = self._coordinates
        
        size = pts.shape[0]
        
        y1 = pts[size-1,1]
        x1 = pts[size-1,0]
        for i in range(size):
            y2 = pts[i,1]
            x2 = pts[i,0]
            if (y1 <= Y < y2) or (y2 <= Y < y1):
                if (x1 + (Y-y1)*(x2 - x1)/(y2 - y1)) > X:
                    ct = not ct
            y1 = y2
            x1 = x2
        return ct
    
        



import numpy


def normaliseVector(a):
    """normalise a (3,) vector or a (n,3) array of vectors"""
    a= numpy.asarray(a)
    mag = numpy.sqrt((a**2).sum(axis=-1))[...,numpy.newaxis]
    return a/mag

def dotprod(a,b):
    """dot-product along last axis"""
    return (a*b).sum(axis=-1)[...,numpy.newaxis]

def Convert_to_SP(input_v, normal_v, E1_vector, E1_amp, E2_amp):
    """
    All inputs are 2D arrays
    """
    cross = numpy.cross
    
    E2_vector = cross(input_v, E1_vector)
    
    v = cross(input_v, normal_v)
    S_vector = numpy.where(numpy.all(v==0, axis=1).reshape(-1,1)*numpy.ones(3), 
                           normaliseVector(E1_vector),
                           normaliseVector(v) )
    
    v = cross(input_v, S_vector)
    P_vector = normaliseVector(v)
    
    S_amp = E1_amp*dotprod(E1_vector,S_vector) + E2_amp*dotprod(E2_vector, S_vector)
                
    P_amp = E1_amp*dotprod(E1_vector,P_vector) + E2_amp*dotprod(E2_vector, P_vector)
    
    return S_amp, P_amp, S_vector, P_vector

def rotation(theta):
    #this is maybe deprecated. added z_rotation() because this was rotating in all three dimensions
    #by the same angle.  why would one want to do that?
    tx=ty=tz = theta
    cos = numpy.cos
    sin = numpy.sin
    Rx = numpy.array([[1,0,0], [0, cos(tx), -sin(tx)], [0, sin(tx), cos(tx)]])
    Ry = numpy.array([[cos(ty), 0, -sin(ty)], [0, 1, 0], [sin(ty), 0, cos(ty)]])
    Rz = numpy.array([[cos(tz), -sin(tz), 0], [sin(tz), cos(tz), 0], [0,0,1]])

    return numpy.dot(Rx, numpy.dot(Ry, Rz))

def z_rotation(theta):
    tz = theta
    cos = numpy.cos
    sin = numpy.sin
    Rz = numpy.array([[cos(tz), -sin(tz), 0], [sin(tz), cos(tz), 0], [0,0,1]])
    return Rz




from raypier.mirrors import PECMirror
from raypier.sources import ParallelRaySource
from raypier.tracer import RayTraceModel

m1 = PECMirror(name="M1",
                centre=(0,-45, 0),
                direction=(0.,-1,1), #(0,y_dir,-small_height),
                diameter=40.,
                thickness=4.)

source = ParallelRaySource(origin=(0,0,0),
                           direction=(0,-1,0),
                           radius=12,
                           number=20)

model = RayTraceModel(sources=[source],
                      optics=[m1])

model.configure_traits()
#model.update = True
#print source.traced_rays


#!/usr/bin/env python

# import pyximport
# pyximport.install(language_level=3)

import sys
# sys.path.append('..')
from raypier import ctracer, cmaterials
import unittest
from math import sqrt
import random
import numpy


class TestRayStructure(unittest.TestCase):
    def test_sizeof_ray(self):
        self.assertEqual(ctracer.get_ray_size(), ctracer.ray_dtype.itemsize)


class TestVectorMathsFunctions(unittest.TestCase):
    def test_set_v(self):
        a = (1., 2., 3.)
        b = ctracer.py_set_v(a)
        self.assertEqual(a,b)

    def test_addvv(self):
        a = [1.,2.,3.]
        b = [2.,3.,4.]
        c = ctracer.addvv(a,b)
        self.assertEqual(c, tuple(ai+bi for ai,bi in zip(a,b)))

    def test_invert(self):
        a = (1.2, 3.4, 5.6)
        b = ctracer.invert(a)
        self.assertEqual(a, tuple(-bi for bi in b))

    def test_subvv(self):
        a = [1.,2.,3.]
        b = [2.,3.,4.]
        c = ctracer.subvv(a,b)
        self.assertEqual(c, tuple(ai-bi for ai,bi in zip(a,b)))

    def test_multvv(self):
        a = [1.,2.,3.]
        b = [2.,3.,4.]
        c = ctracer.multvv(a,b)
        self.assertEqual(c, tuple(ai*bi for ai,bi in zip(a,b)))

    def test_addvs(self):
        a = [1.,2.,3.]
        b = 4.56
        c = ctracer.addvs(a,b)
        self.assertEqual(c, tuple(ai+b for ai in a))

    def test_subvs(self):
        a = [1.,2.,3.]
        b = 4.56
        c = ctracer.subvs(a,b)
        self.assertEqual(c, tuple(ai-b for ai in a))

    def test_multvs(self):
        a = [1.,2.,3.]
        b = 4.56
        c = ctracer.multvs(a,b)
        self.assertEqual(c, tuple(ai*b for ai in a))

    def test_sep(self):
        a = [1., 2., 3.]
        b = [4., 5., 6.]
        c = ctracer.sep(a,b)
        self.assertEqual(c, sqrt(sum((ai-bi)**2 for ai,bi in zip(a,b))))

    def test_dotprod(self):
        a = [1., 2., 3.]
        b = [4., 5., 6.]
        c = ctracer.dotprod(a,b)
        self.assertEqual(c, sum(x*y for x,y in zip(a,b)))

    def test_cross(self):
        a = [1., 2., 3.]
        b = [4., 5., 6.]
        c = ctracer.cross(a,b)
        A = ctracer.dotprod(a,c)
        B = ctracer.dotprod(b,c)
        self.assertEqual(A, 0.0)
        self.assertEqual(B, 0.0)

    def test_map(self):
        a = [3.4, 5.6, 7.8]
        b = ctracer.mag(a)
        self.assertEqual(b, sqrt(sum(x**2 for x in a)))

    def test_norm(self):
        a = [1.1, 2.2, 3.3]
        b = ctracer.norm(a)
        print("B", b)
        self.assertAlmostEqual(1.0, sum(x**2 for x in b))
        self.assertAlmostEqual(0.0, ctracer.mag(ctracer.cross(a,b)))


class TestPECMaterial(unittest.TestCase):
    def test_eval_child(self):
        m = cmaterials.PECMaterial()
        in_ray = ctracer.Ray(origin=(-1,0,-1), direction=(1,0,1))
        out_rays = ctracer.RayCollection(1)
        m.eval_child_ray(in_ray, 1, (0,0,0), (0,0,1), (0,1,0), out_rays)
        out_ray = out_rays[0]
        self.assertEqual(out_ray.direction, (1,0,-1))


class TestPolarisation(unittest.TestCase):
    def test_ellipticity_RHS(self):
        phase = (8.53 + 2.75j)
        in_ray = ctracer.Ray(origin=(-1,0,-1),
                             direction=(1,0,1),
                             E_vector=(1.,2.,3.),
                             E1_amp=phase*(1.0+0.0j)/numpy.sqrt(2),
                             E2_amp=phase*(0.0+1.0j)/numpy.sqrt(2))

        self.assertEqual(in_ray.ellipticity, 1.0)

    def test_ellipticity_LHS(self):
        phase = (1.23 + 3.45j)
        in_ray = ctracer.Ray(origin=(-1,0,-1),
                             direction=(1,0,1),
                             E_vector=(1.,2.,3.),
                             E1_amp=phase*(2.0+0.0j)/numpy.sqrt(2),
                             E2_amp=phase*(0.0-2.0j)/numpy.sqrt(2))

        self.assertEqual(in_ray.ellipticity, -1.0)

    def test_linear(self):
        in_ray = ctracer.Ray(origin=(-1,0,-1),
                             direction=(1,0,1),
                             E_vector=(1.,2.,3.),
                             E1_amp=(1.0+2.0j)/numpy.sqrt(2),
                             E2_amp=(2.0+4.0j)/numpy.sqrt(2))

        self.assertAlmostEqual(in_ray.ellipticity, 0.0)


class TestRay(unittest.TestCase):
    def test_termination(self):
        a = (1,2,3)
        b = (4,5,6)
        c = 2.3
        ray = ctracer.Ray(origin=a, direction=b)
        ray.length = c
        result = tuple(A+c*B for A,B in zip(a,b))
        shift = sum((a-b)**2 for a,b in zip(result, ray.termination))
        self.assertAlmostEqual(shift, 0.0)


class TestRayCollection(unittest.TestCase):
    def test_iteration(self):
        ray = ctracer.Ray(origin=(-1,0,-1), direction=(1,0,1))
        rc = ctracer.RayCollection(10)
        for i in range(6):
            rc.add_ray(ray)
        self.assertEqual(rc.n_rays, 6)
        rays = [r for r in rc]
        self.assertEqual(len(rays), 6)

    def test_iteration2(self):
        ray = ctracer.Ray(origin=(-1,0,-1), direction=(1,0,1))
        rc = ctracer.RayCollection(10)
        for i in range(6):
            rc.add_ray(ray)
        self.assertEqual(rc.n_rays, 6)
        itr = iter(rc)
        self.assertEqual(type(itr), ctracer.RayCollectionIterator)
        for i in range(6):
            r1 = next(itr)
            self.assertEqual(r1.origin, ray.origin)
            self.assertEqual(r1.direction, ray.direction)

    @staticmethod
    def make_ray():
        rnd = random.random
        r = ctracer.Ray(origin=(rnd(), rnd(), rnd()),
                        direction=(rnd(), rnd(), rnd()),
                        E1_amp = (rnd() + 1j*rnd()),
                        E2_amp = (rnd() + 1j*rnd()),
                        refractive_index = (rnd() + 1j*rnd())
                        )
        return r

    @staticmethod
    def compare(ray, row):
        a = tuple(ray.origin)==tuple(row['origin'])
        b = tuple(ray.direction)==tuple(row['direction'])
        return a and b

    def test_copy_to_array(self):
        rc = ctracer.RayCollection(10)
        rays = [self.make_ray() for i in range(7)]
        for ray in rays:
            rc.add_ray(ray)
        data = rc.copy_as_array()
        is_same = [self.compare(ray,row) for ray,row in zip(rays, data)]
        self.assertTrue(all(is_same))

    def test_from_array(self):
        a = numpy.empty(5, dtype=ctracer.ray_dtype)
        rc = ctracer.RayCollection.from_array(a)
        for i in range(len(a)):
            self.assertEqual(tuple(a[i]['origin']), tuple(rc[i].origin))
            self.assertEqual(tuple(a[i]['direction']), tuple(rc[i].direction))

    def test_properties(self):
        rc = ctracer.RayCollection(10)
        rays = [self.make_ray() for i in range(7)]
        for ray in rays:
            rc.add_ray(ray)
        data = rc.copy_as_array()
        self.assertTrue( numpy.alltrue( rc.E1_amp==data['E1_amp'] ) )
        self.assertTrue( numpy.alltrue( rc.E2_amp==data['E2_amp'] ) )
        self.assertTrue( numpy.alltrue( rc.refractive_index==data['refractive_index'] ) )


class AnOwner(object):
    def __init__(self, **kwds):
        self.__dict__.update(kwds)


def sep(a,b):
    return sum((A-B)**2 for A,B in zip(a,b))


class TestTransform(unittest.TestCase):
    """
    def test_trans(self):
        print "loading tvtk"
        from tvtk.api import tvtk
        print "complete"
        t = tvtk.Transform()
        t.rotate_x(10)
        t.rotate_y(15)
        t.rotate_z(20)
        t.translate(3,4,5)
        o = AnOwner(transform=t)
        fl = ctracer.FaceList(owner=o)
        fl.sync_transforms()
        print fl.transform.rotation
        print fl.transform.translation
        pt = (0.2, 0.4, 0.6)
        pt2 = t.transform_double_point(*pt)
        pt3 = ctracer.transform(fl.transform, pt)
        print pt, pt2, pt3
        self.assertEquals(sep(pt2,pt3), 0.0)
    """


class TestTraceSegment(unittest.TestCase):
    def test_trace_segment(self):
        from raypier import cfaces

        o = AnOwner(diameter=5.5, offset=6.6)
        c = cfaces.CircularFace(owner=o)

        rays = ctracer.RayCollection(10)
        in_ray = ctracer.Ray(origin=(-1,0,-1), direction=(1,0,1))
        rays.add_ray(in_ray)

        self.assertEqual(rays.n_rays, 1)

        face_set = ctracer.FaceList()
        face_set.faces = [c]
        all_faces = [c]

        out_rays = ctracer.trace_segment(rays, [face_set], all_faces)

        self.assertEqual(out_rays.n_rays, 1)

        out_ray = out_rays[0]

        self.assertEqual(out_ray.direction, (1,0,-1))


class TestInterfaceMaterial(unittest.TestCase):
    def test_wavelengths(self):
        m = ctracer.InterfaceMaterial()
        a = [1.,2.,3.]
        m.wavelengths = numpy.array(a)
        b = list(m.wavelengths)

        self.assertEqual(a,b)

if __name__=="__main__":
    unittest.main()



"""
A module for Results subclasses
"""


from traits.api import Instance, Float, on_trait_change,\
            Button, DictStrFloat, Str

from traitsui.api import View, Item, DropEditor

from raypier.bases import Result, Traceable
from raypier.sources import BaseRaySource
#from raypier.tracer import RayTraceModel
from raypier.core.ctracer import Face
from raypier.dispersion import FusedSilica

from traitsui.editors.drop_editor import DropEditor
from traitsui.editors.title_editor import TitleEditor

import numpy
import itertools
import traceback


class TargetResult(Result):
    """A base-class for result calculated from rays terminating on a chosen face"""
    source = Instance(BaseRaySource)
    target = Instance(Face)
    name = Str("TargetResult")
    result = Float(label="Result", transient=True)
    
    traits_view = View(Item("name", style="readonly", editor=TitleEditor()), 
                       Item('result', style="readonly"),
                       Item('source', editor=DropEditor()),
                       Item('target', editor=DropEditor()),
                       title="Optical path length",
                       resizable=True,
                       )
    
    def calc_result(self, tracer):
        try:
            self._calc_result()
        except:
            traceback.print_exc()
        
    def _calc_result(self):
        raise NotImplementedError()
        
    @on_trait_change("source, target")
    def update(self):
        if self.target is None:
            return
        if self.source is None:
            return
        if self.source.traced_rays:
            try:
                self._calc_result()
            except:
                traceback.print_exc()
        

class MeanOpticalPathLength(TargetResult):
    name = "Mean optical path length"
    abstract = False
    
    result = Float(label="Path length", transient=True)
    
    traits_view = View(Item('result', style="readonly"),
                       Item('source', editor=DropEditor()),
                       Item('target', editor=DropEditor()),
                       title="Optical path length",
                       resizable=True,
                       )
    
    @on_trait_change("source, target")
    def update(self):
        if self.target is None:
            return
        if self.source is None:
            return
        if self.source.traced_rays:
            self._calc_result()
    
    def _calc_result(self):
        all_rays = [r.copy_as_array() for r in reversed(self.source.traced_rays)]
        idx = self.target.idx
        last = all_rays[0]
        selected_idx = numpy.argwhere(last['end_face_idx']==idx).ravel()
        total = numpy.zeros(len(selected_idx), 'd')
        for ray in all_rays:
            selected = ray[selected_idx]
            total += selected['length'] * selected['refractive_index'].real
            selected_idx = selected['parent_idx']
        self.result = total.mean()
        
        
def evaluate_phase(all_wavelengths, traced_rays, target_face,
                   glass_length=0.0, glass_dispersion=FusedSilica()):
    """
    
    all_wavelengths - a numpy array giving the wavelengths, in microns
    traced_rays - the list of RayCollections from the source object after tracing
    target_face - the Face object at which we terminate the tracing
    glass_length - an additional length of glass added to the computation
    glass_dispersion - the DispersionCurve for the extra glass (default Fused Silica)
    
    returns - (freq, phase) #freq in THz
    """
    c = 2.99792458e8 * 1e-9 #convert to mm/ps
    idx = target_face.idx
    last = traced_rays[-1].copy_as_array() #all_rays[0]
    selected_idx = numpy.argwhere(last['end_face_idx']==idx).ravel()
    wavelengths = all_wavelengths[last['wavelength_idx'][selected_idx]]
    sort_idx = numpy.argsort(wavelengths)[::-1]
    wavelengths = wavelengths[sort_idx]
    selected_idx = selected_idx[sort_idx]
    
    idx = len(selected_idx)//2
    phase = last['phase'][selected_idx].copy()
    phase -= phase.mean()
    
    total = last['accumulated_path'][selected_idx]

    if len(total) < 6:
        raise ValueError("Not enough rays to evaluate 2nd and 3rd derivatives")
    ave_path = total.mean()
    total -= ave_path #because we don't care about the overall offset
    f = 1000.0*c/wavelengths #in THz
    
    n_fs = glass_dispersion.evaluate_n(wavelengths).real
    fs_total = n_fs*glass_length*1000.0 #convert path length to mm
    fs_total -= fs_total.mean()
    total += fs_total
    
    phase += total*f*((2*numpy.pi)/c)
    return (f, phase)
    

class GroupVelocityDispersion(MeanOpticalPathLength):
    name = "Group Velocity Dispersion"
    wavelength = Float(label="Wavelength /um", transient=True)
    result = Float(label="GDD /fs^2", transient=True)
    tod = Float(label="TOD", transient=True)
    glass_path = Float(0.0)
    
    _fs = Instance(FusedSilica, ())
    
    traits_view = View(Item("wavelength", style="readonly"),
                       Item('result', style="readonly"),
                       Item('tod', style="readonly"),
                       Item("glass_path"),
                       Item('source', editor=DropEditor()),
                       Item('target', editor=DropEditor()),
                       title="Optical path length",
                       resizable=True,
                       )
    
    @on_trait_change("glass_path")
    def _glass_path_update(self):
        self.update()
    
    def _calc_result(self):
        all_wavelengths = numpy.asarray(self.source.wavelength_list)
        traced_rays = self.source.traced_rays
        target_face = self.target
        glass_length = self.glass_path
        glass_dispersion = self._fs
        f, phase = evaluate_phase(all_wavelengths, traced_rays, target_face, glass_length, glass_dispersion)
        
        omega = f*(2*numpy.pi)
        dw = numpy.diff(omega)
        dw_mean = dw.mean()
        #dw_sd = numpy.std(dw)
        #print("uniformity in f:", dw_sd, dw_mean)
        #print "FREQ:", f
        second_deriv = numpy.diff(phase,2)/(dw_mean**2)
        third_deriv = numpy.diff(phase,3)/(dw_mean**3)
        second_deriv *= 1e6 #convert to fs^2
        third_deriv *= 1e9 #convert to fs^3
        
        c = 2.99792458e8 * 1e-9 #convert to mm/ps
        
        idx = target_face.idx
        self.result = second_deriv[idx]
        self.wavelength = c*1000.0/numpy.median(f) #convert to nm
        self.tod = third_deriv[idx]
        
        
class FocalPoint(TargetResult):
    pass #TODO


class Divergence(TargetResult):
    abstract = False
    name = "Beam Divergence"
    
    traits_view = View(Item('result', style="readonly"),
                       Item('source', editor=DropEditor()),
                       Item('target', editor=DropEditor()),
                       title="Optical path length",
                       resizable=True,
                       )
    
    def _calc_result(self):
        all_rays = [r.copy_as_array() for r in reversed(self.source.traced_rays)]
        idx = self.target.idx
        last = all_rays[0]
        selected_idx = numpy.argwhere(last['end_face_idx']==idx).ravel()
        selected_rays = last[selected_idx]
        directions = selected_rays['direction']
        ave_direction = directions.mean(axis=0,keepdims=True)
        cp_vectors = numpy.cross(ave_direction,directions)
        rms_cp = (cp_vectors**2).sum(axis=-1).mean()
        self.result = rms_cp 
        
        
def get_total_intersections(raysList, face):
    all_rays = itertools.chain(*raysList)
    '''#debug
    print face.idx, " all_rays: "
    for ray in all_rays:
    print ray.end_face_idx  #'''
    idx = face.idx
    return sum(1 for ray in all_rays if ray.end_face_idx==idx)


def get_total_power(raysList, face):
    all_rays = itertools.chain(*raysList)
    idx = face.idx
    return sum(ray.power for ray in all_rays if ray.end_face_idx==idx)


class RayPaths(Result):
    ''' keep a tally of how many rays follow a particular path from face to face.  result 
    is a dictionary where the keys are lists of face indicies like [0,3,2,3,463772762] and 
    the values are the percentage of rays that encountered faces in that order.  The Huge
    face_idx is infinity and is always the last "face" 

    check result.keys() to see which paths were taken.'''

    name = "ray paths"
    result = DictStrFloat(label="Dictionary of ray paths")

    _tracer = Instance("raypier.tracer.RayTraceModel") #to cache the tracer instance
    
    traits_view = View(Item('result', style="readonly"),
                       title="Paths taken by rays",
                       resizable=True,
                       )

    @on_trait_change("_tracer")
    def update(self):
        if self._tracer is not None:
            self._calc_result()
        
    def calc_result(self, tracer):
        self._tracer = tracer
        self._calc_result()
    
    def _calc_result(self):
                
        result = {}    
        for source in self._tracer.sources:
            #a list of RayCollection instances
            rays = source.get_ray_list_by_id()
            #print "how many rays?",len(rays)
            for ray in rays:
                path = []
                for segment in ray:
                    if segment['end_face_idx'] < 4294967200:
                        c = segment['end_face_idx']
                    else: c = numpy.Infinity
                    path.append(c)
                #get_ray_list_by_id goes from last ray to first, which isn't as pretty as forwards in time.
                path.reverse()      
                if str(path) in list(result.keys()):
                    result[str(path)] += 1
                else:
                    result[str(path)] = 1
        total = float(sum(result.values()))
        for key in list(result.keys()):
            result[key] = result[key]/total
        self.result = result   


class Ratio(Result):
    name = "a ratio"
    abstract = False
    nominator = Instance(Face)
    denominator = Instance(Face)
    
    #because I always get them the wrong way round!
    switch_faces = Button("Switch faces")
    
    result = Float(label="Ratio of intersections")
    
    _tracer = Instance("raypier.tracer.RayTraceModel") #to cache the tracer instance
    
    traits_view = View(Item('result', style="readonly"),
                       Item('nominator', editor=DropEditor()),
                       Item('denominator', editor=DropEditor()),
                       Item('switch_faces', show_label=False),
                       title="Face intersection ratio",
                       resizable=True,
                       )
    
    def _switch_faces_changed(self):
        self.nominator, self.denominator = \
            self.denominator, self.nominator
        self._calc_result()
    
    @on_trait_change("nominator, denominator, _tracer")
    def update(self):
        if self._tracer is not None:
            self._calc_result()
        
    def calc_result(self, tracer):
        self._tracer = tracer
        self._calc_result()
    
    def _calc_result(self):
        nom = self.nominator
        denom = self.denominator
        if not all((nom, denom)):
            return
        
        nom_count = 0
        denom_count = 0
        #just sum result from multiple sources?
        #maybe a dictionary or something would be better?
        for source in self._tracer.sources:
            #a list of RayCollection instances
            raysList = source.traced_rays
        
            nom_count = nom_count + get_total_intersections(raysList, nom)
            denom_count = denom_count + get_total_intersections(raysList, denom)
	#print "nom and denom counts", nom_count, denom_count
        try:
            self.result = float(nom_count)/float(denom_count)
        except ZeroDivisionError:
            self.result = numpy.Infinity
        
        
class IncidentPower(Ratio):
    name = "Incident power"
    abstract = False
    object = Instance(Traceable)
    
    traits_view = View(Item('result', style="readonly"),
                       Item('object', editor=DropEditor()),
                       title="Incident power on object",
                       resizable=True,
                       )
    
    def _calc_result(self):
        nom = self.object
        if not nom:
            return
        
        nom_count = 0.0
        power_in = 0.0
        #just sum result from multiple sources?
        #maybe a dictionary or something would be better?
        for source in self._tracer.sources:
            #a list of RayCollection instances
            raysList = source.traced_rays
            input = source.traced_rays[0]
            E1_amp = input.E1_amp
            E2_amp = input.E2_amp
            n = input.refractive_index.real
            power_in += (n*E1_amp*E1_amp.conjugate()).real.sum() + (n*E2_amp*E2_amp.conjugate()).real.sum()
            
            for f in nom.faces.faces:
                for rc in raysList:
                    rca = rc.copy_as_array()
                    selected = rca[rca['end_face_idx']==f.idx]
                    E1_amp = selected['E1_amp']
                    E2_amp = selected['E2_amp']
                    n = selected['refractive_index'].real
                    nom_count += (n*E1_amp*E1_amp.conjugate()).real.sum() +\
                                (n*E2_amp*E2_amp.conjugate()).real.sum()
            
    #print "nom and denom counts", nom_count, denom_count
        try:
            self.result = float(nom_count)/float(power_in)
        except ZeroDivisionError:
            self.result = numpy.Infinity



#    Copyright 2009, Teraview Ltd., Bryan Cole
#
#    This file is part of Raypier.
#
#    Raypier is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import numpy
from tvtk.api import tvtk
from traits.api import Range as _Range, Tuple as _Tuple,\
            BaseTuple
from traitsui.api import TupleEditor

### For backwards compatbility
from .core.utils import normaliseVector, dotprod, Convert_to_SP, rotation, z_rotation


class EditorTraits(object):
    def get_editor(self, *args, **kwds):
        e = super(EditorTraits, self).get_editor(*args, **kwds)
        editor_t = {'auto_set':False,
                    'enter_set':True}
        metadata = self._metadata
        if 'editor_traits' in metadata:
            editor_t.update(metadata['editor_traits'])
        e.set(**editor_t)
        return e

class Range(EditorTraits, _Range):
    pass
    
    
class Tuple(EditorTraits, _Tuple):
    pass


class UnitVectorTrait(EditorTraits, BaseTuple):
    def validate(self, object, name, value):
        value = super(UnitVectorTrait, self).validate(object,name,value)
        mag = numpy.sqrt(sum(a**2 for a in value))
        return tuple(float(a/mag) for a in value)


TupleVector = Tuple((0.,0.,0.), editor_traits={'cols':3,
                                             'labels':['x','y','z']})
                                             
UnitTupleVector = UnitVectorTrait((0.,0.,1.), editor_traits={'cols':3,
                                             'labels':['x','y','z']})



def transformPoints(t, pts):
    """Apply a vtkTransform to a numpy array of points"""
    out = tvtk.Points()
    t.transform_points(pts, out)
    return numpy.asarray(out)

def transformVectors(t, pts):
    """Apply a vtkTransform to a numpy array of vectors.
    I.e. ignore the translation component"""
    out = tvtk.DoubleArray(number_of_components=3)
    t.transform_vectors(pts, out)
    return numpy.asarray(out)

def transformNormals(t, pts):
    """Apply a vtkTransform to a numpy array of normals.
    I.e. ignore the translation component and normalise
    the resulting magnitudes"""
    out = tvtk.DoubleArray(number_of_components=3)
    t.transform_normals(pts, out)
    return numpy.asarray(out)




#!/usr/bin/env python
from _operator import invert

cdef extern from "math.h":
    double sqrt(double arg)

cdef extern from "float.h":
    #double INFINITY
    double DBL_MAX

cdef:
    double INF
    int NPARA=6


from libc.stdlib cimport malloc, free

############################################
### C type declarations for internal use ###
############################################

cdef struct vector_t:
    double x,y,z

cdef struct orientation_t:
    vector_t normal, tangent


IF UNAME_SYSNAME == "Windows":
    ctypedef double complex complex_t
ELSE:
    ctypedef double complex complex_t
#     cdef struct complex_t:
#         double real
#         double imag
        
cdef:
    public unsigned int REFL_RAY=1<<0
    public unsigned int GAUSSLET=1<<1
    public unsigned int PARABASAL=1<<2


cdef packed struct ray_t:
    #vectors
    vector_t origin, direction, normal, E_vector
    #complex attribs
    complex_t refractive_index, E1_amp, E2_amp
    #simple attribs
    double length, phase, accumulated_path
    #reference ids to related objects
    unsigned int wavelength_idx, parent_idx, end_face_idx, ray_type_id
    
    ##objects
    #object face, end_face, child_refl, child_trans
    
### Structure for parabasal rays. These only carry geometric information.
cdef packed struct para_t:
    #vectors
    vector_t origin, direction, normal
    #simple attribs
    double length

    
### A Gausslet combines a base-ray with 6 parabasal rays
cdef packed struct gausslet_t:
    ray_t base_ray
    para_t para[6]
    

cdef struct transform_t:
    double m00, m01, m02, m10, m11, m12, m20, m21, m22
    double tx, ty, tz

cdef struct ray_pair_t:
    ray_t trans, refln

##############################
### Vector maths functions ###
##############################

cdef  vector_t transform_c(transform_t t, vector_t p) nogil
cdef  vector_t rotate_c(transform_t t, vector_t p) nogil
cdef  vector_t set_v(object O)
cdef  double sep_(vector_t p1, vector_t p2) nogil
cdef  vector_t multvv_(vector_t a, vector_t b) nogil
cdef  vector_t multvs_(vector_t a, double b) nogil
cdef  vector_t addvv_(vector_t a, vector_t b) nogil
cdef  vector_t addvs_(vector_t a, double b) nogil
cdef  vector_t subvv_(vector_t a, vector_t b) nogil
cdef  vector_t subvs_(vector_t a, double b) nogil
cdef  double dotprod_(vector_t a, vector_t b) nogil
cdef  vector_t cross_(vector_t a, vector_t b) nogil
cdef  vector_t norm_(vector_t a) nogil
cdef  double mag_(vector_t a) nogil
cdef  double mag_sq_(vector_t a) nogil
cdef  vector_t invert_(vector_t v) nogil


##################################
### Python extension types
##################################

cdef class Transform:
    cdef transform_t trans
    

cdef class Ray:
    cdef:
        ray_t ray
        double max_length
    

cdef class ParabasalRay:
    cdef:
        para_t ray
        double max_length
    
    
cdef class Gausslet:
    cdef gausslet_t gausslet
    
    
cdef class RayArrayView:
    cdef void set_ray_c(self, unsigned long i, ray_t ray)
    cdef ray_t get_ray_c(self, unsigned long i)
    cdef unsigned long get_n_rays(self)
    

cdef class RayCollection(RayArrayView):
    cdef: 
        ray_t *rays
        readonly unsigned long n_rays, max_size
        RayCollection _parent
        double[:] _wavelengths
        
        int[:,:] _neighbours
        double _mtime        

    cdef add_ray_c(self, ray_t r)
    cdef void reset_length_c(self, double max_length)
    
    cdef double get_mtime(self, unsigned long guard)
    
    cdef void _eval_neighbours(self, int[:,:] pnb)
    
    
cdef class GaussletCollection:
    cdef: 
        gausslet_t *rays
        readonly unsigned long n_rays, max_size
        GaussletCollection _parent
        double[:] _wavelengths

    cdef add_gausslet_c(self, gausslet_t r)
    cdef void extend_c(self, GaussletCollection gc)
    cdef void reset_length_c(self, double max_length)
    
    
cdef class GaussletBaseRayView(RayArrayView):
    cdef:
        GaussletCollection owner
    

cdef class RayCollectionIterator:
    cdef:
        RayCollection rays
        unsigned int counter
        
        
cdef class GaussletCollectionIterator:
    cdef:
        GaussletCollection rays
        unsigned int counter


cdef class InterfaceMaterial(object):
    """Abstract base class for objects describing
    the materials characterics of a Face
    """
    cdef double[:] _wavelengths

    cdef void eval_child_ray_c(self, ray_t *old_ray,
                            unsigned int ray_idx,
                            vector_t point,
                            orientation_t orient,
                            RayCollection new_rays)
    
    cdef para_t eval_parabasal_ray_c(self, ray_t *base_ray,  
                                     vector_t direction, #incoming ray direction
                                   vector_t point, #position of intercept
                                   orientation_t orient,
                                   unsigned int ray_type_id, #indicates if it's a transmitted or reflected ray 
                                   )
    
    cdef void eval_decomposed_rays_c(self, GaussletCollection child_rays)

    cdef on_set_wavelengths(self)


cdef class Distortion:
    cdef vector_t z_offset_and_gradient_c(self, double x, double y) nogil
    cdef double z_offset_c(self, double x, double y) nogil
    
    
cdef class Shape:
    cdef bint point_inside_c(self, double x, double y)


cdef class Face(object):
    cdef:
        public object owner
        public char *name
        public double tolerance
        public int idx #index in the global face list
        public double max_length
        public InterfaceMaterial material
        public short int invert_normal
        public unsigned int count    

    cdef double intersect_c(self, vector_t p1, vector_t p2, int is_base_ray)

    cdef vector_t compute_normal_c(self, vector_t p)
    cdef vector_t compute_tangent_c(self, vector_t p)


cdef class FaceList(object):
    """A group of faces which share a transform"""
    cdef transform_t trans
    cdef transform_t inv_trans
    cdef public list faces
    cdef public object owner

    cpdef void sync_transforms(self)
    cdef int intersect_c(self, ray_t *ray, vector_t end_point)
    cdef int intersect_para_c(self, para_t *ray, vector_t ray_end, Face face)
    cdef orientation_t compute_orientation_c(self, Face face, vector_t point)


##################################
### Python module functions
##################################

cdef RayCollection trace_segment_c(RayCollection rays,
                                    list face_sets,
                                    list all_faces,
                                    list decomp_faces,
                                    float max_length)

cdef GaussletCollection trace_gausslet_c(GaussletCollection gausslets, 
                                    list face_sets, 
                                    list all_faces,
                                    list decomp_faces,
                                    double max_length)

cdef double ray_power_(ray_t ray)




import unittest
import numpy

from raypier.core.ctracer import GaussletCollection, gausslet_dtype, ray_dtype, RayCollection


class TestGaussletCollectionFromRays(unittest.TestCase):
    def test_from_rays(self):
        data = numpy.zeros(3, dtype=ray_dtype)
        data['origin'] = [[1.,2.,3.]]
        data['direction'] = [[4.,5.,6.]]
        data['length'] = [7.0,]
        
        gc = GaussletCollection.from_rays(data)
        rc = RayCollection.from_array(data)
        
        agc = gc[0]
        ar = rc[0]
        
        self.assertEqual(agc.base_ray.origin, ar.origin)
        self.assertEqual(agc.base_ray.direction, ar.direction)
        self.assertEqual(agc.base_ray.normal, ar.normal)
        self.assertEqual(agc.base_ray.length, ar.length)
        for i in range(6):
            paras = agc.parabasal_rays
            self.assertEqual(paras[i].origin, ar.origin)
            self.assertEqual(paras[i].direction, ar.direction)
            self.assertEqual(paras[i].length, ar.length)
